# Supports VPTQ compression, see https://arxiv.org/abs/2409.17066

import math
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn.functional as F
from torch.nn.parameter import Parameter

from aphrodite import _custom_ops as ops
from aphrodite.distributed import get_tensor_model_parallel_rank
from aphrodite.modeling.layers.linear import LinearBase, LinearMethodBase
from aphrodite.modeling.utils import set_weight_attrs
from aphrodite.quantization import QuantizationMethods
from aphrodite.quantization.base_config import QuantizationConfig


class MetaData:

    def __init__(self):
        self.num_codebooks = 1
        self.num_centroids = 0
        self.num_res_centroids = 0
        self.vector_len = 0
        self.group_size = 0
        self.output_size = 0


# unpack the packed tensor to get the indices and residual indices
def unpack_index_tensor(
    pack_tensor: torch.Tensor,
    index_bits: int,
    num_elements: int,
    res_bits: int = 0,
    num_res_elements: int = 0,
    index_dtype: torch.dtype = torch.uint16,
    as_dtype: torch.dtype = torch.int32,
) -> torch.Tensor:
    total_bits = index_bits + res_bits
    wf = torch.arange(0, 32, 1, device=pack_tensor.device).view(1, 1, 1, -1)
    out = torch.bitwise_right_shift(torch.unsqueeze(pack_tensor, -1), wf)
    torch.bitwise_and(out, 1, out=out)
    pad_size = (pack_tensor.shape[-1] * 32) % (index_bits * num_elements +
                                               res_bits * num_res_elements)
    out = out.reshape(*pack_tensor.shape[:-1], -1)
    if pad_size > 0:
        out = out[..., :-pad_size]
    out = out.reshape(*pack_tensor.shape[:-1], -1, total_bits)
    wf1 = torch.arange(0, total_bits, 1,
                       device=pack_tensor.device).view(1, 1, 1, -1)
    out = torch.bitwise_left_shift(out, wf1).sum(dim=-1)

    unpack_indice = out.to(torch.uint64).view(torch.int64)

    indices = (unpack_indice & ((1 << index_bits) - 1)).view(torch.uint64).to(
        torch.int64)

    # indices = indices.squeeze()

    if res_bits > 0:
        res_indices = ((unpack_indice >> index_bits) &
                       ((1 << index_bits) - 1)).view(torch.uint64).to(
                           torch.int64)
        # res_indices = res_indices.squeeze()
    else:
        res_indices = None

    return indices, res_indices


# dequantize the weight from the quantization codes
def dequantize_weight(
        indices: torch.
    IntTensor,  #  [num_out_groups, num_in_groups, num_codebooks]
        codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        res_codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        weight_scale: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        weight_bias: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        perm: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        metadata: MetaData) -> torch.Tensor:
    """
    Decode float weights from quantization codes. Differentiable.
    :param codes: tensor of integer quantization codes, shape 
        [*dims, num_out_groups, num_in_groups, num_codebooks]
    :param codebooks: tensor of vectors for each quantization code, 
        [num_codebooks, codebook_size, out_group_size, in_group_size]
    :param scales: weight will be multiplied by this factor, must be 
        broadcastble with 
        [*dims, out_groups, num_in_groups, out_group_size, in_group_size]
    :return: reconstructed weight tensor of shape 
        [*dims, num_in_groups*group_size]
    """
    output_size = metadata.output_size
    num_codebooks = indices.shape[0]
    num_centroids = metadata.num_centroids
    num_res_centroids = metadata.num_res_centroids
    vector_len = metadata.vector_len
    group_size = metadata.group_size

    codebooks = codebooks.view(num_codebooks, num_centroids, vector_len)
    res_codebooks = res_codebooks.view(
        num_codebooks, num_res_centroids,
        vector_len) if num_res_centroids > 0 else None
    index_bits = math.ceil(math.log2(num_centroids))
    enable_residual = num_res_centroids > 0
    index_res_bits = math.ceil(
        math.log2(num_res_centroids)) if enable_residual else 0

    # print(f"indices shape: {indices.shape}")
    indices, res_indices = unpack_index_tensor(
        pack_tensor=indices,
        index_bits=index_bits,
        num_elements=group_size,
        res_bits=index_res_bits,
        num_res_elements=group_size,
        index_dtype=torch.uint16,
    )

    indices = indices.unsqueeze(-1).expand(-1, -1, -1, vector_len)
    indices = indices.reshape(num_codebooks, -1, vector_len)
    selected_centroids = torch.gather(codebooks, 1, indices)
    selected_centroids = selected_centroids.view(num_codebooks, -1, group_size,
                                                 vector_len)
    selected_centroids = selected_centroids.permute(0, 1, 3, 2)

    qweight = selected_centroids.reshape(num_codebooks, -1, group_size)
    qweight = qweight.permute(1, 0, 2)
    qweight = qweight.reshape(-1, num_codebooks * group_size)

    if enable_residual:
        res_codebooks = res_codebooks.view(num_codebooks, num_res_centroids,
                                           vector_len)
        res_indices = res_indices.unsqueeze(-1).expand(-1, -1, -1, vector_len)
        res_indices = res_indices.reshape(num_codebooks, -1, vector_len)
        selected_res_centroids = torch.gather(res_codebooks, 1, res_indices)
        selected_res_centroids = selected_res_centroids.reshape(
            num_codebooks, -1, group_size, vector_len)
        selected_res_centroids = selected_res_centroids.permute(0, 1, 3, 2)
        qweight = qweight + (selected_res_centroids.reshape(
            num_codebooks, -1, group_size).permute(1, 0, 2).reshape(
                -1, num_codebooks * group_size))

    padding = -output_size % vector_len
    if padding > 0:
        qweight = qweight[:-padding, :]

    enable_perm = perm is not None
    if enable_perm:
        invert_perm = torch.argsort(perm.view(torch.uint16).to(torch.int64))
        qweight = qweight[:, invert_perm]

    enable_norm = weight_scale is not None
    if enable_norm:
        qweight = qweight * weight_scale
        qweight = qweight + weight_bias

    return qweight


# do the quantized matmul in a generic way, it's quite slow
def generic_dequantize_gemm(
        input: torch.Tensor,  #  [..., in_features]
        indices: torch.
    IntTensor,  #  [num_out_groups, num_in_groups, num_codebooks]
        codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        res_codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        weight_scale: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        weight_bias: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        perm: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        bias: Optional[torch.Tensor],
        metadata: MetaData) -> torch.Tensor:
    dequantized_weight = dequantize_weight(
        indices,
        codebooks,
        res_codebooks,
        weight_scale,
        weight_bias,
        perm,
        metadata,
    )
    return F.linear(input, dequantized_weight, bias)


# call the optimized version of the dequantized matmul
def optimized_dequantize_gemm(
        input: torch.Tensor,  #  [..., in_features]
        indices: torch.
    IntTensor,  #  [num_out_groups, num_in_groups, num_codebooks]
        codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        res_codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        weight_scale: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        weight_bias: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        perm: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        bias: Optional[torch.Tensor],
        metadata: MetaData) -> torch.Tensor:
    codebooks = codebooks.view(metadata.num_codebooks, metadata.num_centroids,
                               metadata.vector_len)
    res_codebooks = res_codebooks.view(
        metadata.num_codebooks, metadata.num_res_centroids,
        metadata.vector_len) if metadata.num_res_centroids > 0 else None
    if input.numel() // input.shape[-1] < 3:
        return ops.vptq_gemm(
            input, indices, codebooks, weight_scale, weight_bias,
            [metadata.vector_len, weight_scale.shape[0], metadata.output_size],
            None, res_codebooks, None, None, perm, bias)
    if perm is None:
        invert_perm = None
    else:
        invert_perm = torch.argsort(perm.view(torch.uint16).to(
            torch.int64)).to(torch.uint16).view(torch.int16)
    dequantized_weight = ops.vptq_dequant(
        indices, codebooks, weight_scale, weight_bias,
        [metadata.vector_len, weight_scale.shape[0], metadata.output_size],
        None, res_codebooks, None, None, invert_perm)
    return F.linear(input, dequantized_weight, bias)


# Handle QKV projection and gate-up projection
# we will do Q K V separately
def merged_dequantize_gemm(
        input: torch.Tensor,  #  [..., in_features]
        indices: torch.
    IntTensor,  #  [num_out_groups, num_in_groups, num_codebooks]
        codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        res_codebooks: torch.
    Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        weight_scale: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        weight_bias: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        perm: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        output_partition_sizes: List[int],
        bias: Optional[torch.Tensor],
        metadata: MetaData) -> torch.Tensor:
    output_shape = input.shape[:-1] + (sum(output_partition_sizes), )
    output = torch.empty(output_shape, dtype=input.dtype, device=input.device)

    indice_sizes = getattr(indices, "shard_sizes", [])
    output_extra_offsets = getattr(indices, "output_offset", [])
    num_codebooks = indices.shape[0]

    tp_rank = get_tensor_model_parallel_rank()
    input_size = input.shape[-1]
    input_offset = 0
    indice_offset = 0
    output_offset = 0
    codebooks_offset = 0

    num_linears = len(output_partition_sizes)
    for linear_idx, output_size, indice_size in zip(range(num_linears),
                                                    output_partition_sizes,
                                                    indice_sizes):
        metadata.output_size = output_size
        if len(output_extra_offsets) > 1:
            metadata.output_size = output_size + output_extra_offsets[tp_rank][
                linear_idx]
        shard_output = optimized_dequantize_gemm(
            input, indices.narrow(1, indice_offset, indice_size),
            codebooks.narrow(0, codebooks_offset, num_codebooks),
            res_codebooks.narrow(0, codebooks_offset, num_codebooks),
            weight_scale.narrow(0, input_offset, input_size),
            weight_bias.narrow(0, input_offset, input_size),
            perm.narrow(0, input_offset, input_size) if perm is not None else
            None, bias if bias is None else bias.narrow(
                0, output_offset, output_size), metadata)

        output_slice = output.narrow(-1, output_offset, output_size)
        if tp_rank > 0 and len(output_extra_offsets) > tp_rank:
            shard_output = shard_output.narrow(
                -1, output_extra_offsets[tp_rank][linear_idx], output_size)
        assert (output_slice.shape == shard_output.shape)
        output_slice.copy_(shard_output)
        output_offset += output_size
        indice_offset += indice_size
        codebooks_offset += num_codebooks
        input_offset += input_size
    return output


class VPTQConfig(QuantizationConfig):
    """Config class for VPTQ.
    Reference: https://github.com/microsoft/VPTQ
    """

    def __init__(
        self,
        config_for_layers: Dict[str, Dict[str, Any]],
        shared_layer_config: Dict[str, Dict[str, Any]],
    ) -> None:
        super().__init__()
        self.config_for_layers = config_for_layers
        self.shared_layer_config = shared_layer_config

    def __repr__(self) -> str:
        return (f"VPTQConfig(config_for_layers={self.config_for_layers}, "
                f"shared_layer_config={self.shared_layer_config})")

    @classmethod
    def get_name(cls) -> QuantizationMethods:
        return "vptq"

    @classmethod
    def get_supported_act_dtypes(cls) -> List[torch.dtype]:
        return [torch.half, torch.bfloat16]

    @classmethod
    def get_min_capability(cls) -> int:
        return 70

    @classmethod
    def get_config_filenames(cls) -> List[str]:
        return []  # no extra configs.

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "VPTQConfig":
        config_for_layers: Dict[str, Any] = {}
        shared_layer_config: Dict[str, Any] = {}
        if "config_for_layers" in config:
            config_for_layers = cls.get_from_keys(config,
                                                  ["config_for_layers"])
        if "shared_layer_config" in config:
            shared_layer_config = cls.get_from_keys(config,
                                                    ["shared_layer_config"])
        assert len(config_for_layers) > 0 or len(shared_layer_config) > 0, \
            "VPTQConfig must have at least one of 'config_for_layers'\
             or 'shared_layer_config'"

        return cls(config_for_layers, shared_layer_config)

    def get_config_for_key(self, prefix:str, key:str):
        merged_name = '.'.join([prefix,key])
        if merged_name in self.config_for_layers:
            return self.config_for_layers[merged_name]
        elif key in self.shared_layer_config:
            return self.shared_layer_config[key]
        else:
            raise ValueError(f"Cannot find config for ({prefix}, {key})")

    def get_quant_method(self, layer: torch.nn.Module,
                         prefix: str) -> Optional["VPTQLinearMethod"]:
        if isinstance(layer, LinearBase):
            linear_name = prefix.split(".")[-1]
            base_name = prefix[:prefix.rfind('.')]
            if linear_name == "qkv_proj":
                quant_config = {
                    "q_proj": self.get_config_for_key(base_name, "q_proj"),
                    "k_proj": self.get_config_for_key(base_name, "k_proj"),
                    "v_proj": self.get_config_for_key(base_name, "v_proj"),
                }
            elif linear_name == "gate_up_proj":
                quant_config = {
                    "gate_proj": self.get_config_for_key(base_name,
                                                         "gate_proj"),
                    "up_proj": self.get_config_for_key(base_name, "up_proj"),
                }
            else:
                quant_config = self.get_config_for_key(base_name, linear_name)
            return VPTQLinearMethod(quant_config)
        return None

    def get_scaled_act_names(self) -> List[str]:
        return []


class VPTQLinearMethod(LinearMethodBase):
    """Linear method for VPTQ.
    Args:
        quant_config: The VPTQ quantization config.
    """

    def __init__(self, quant_config: Dict[str, Any]):
        self.quant_config = quant_config

    @staticmethod
    def quantized_weight_loader(
            indice_sizes,
            narrow_dim=1):  # specific for layer.indices/weight_scale&bias

        def wrap_weight_loader(param: torch.nn.Parameter,
                               loaded_weight: torch.Tensor,
                               loaded_shard_id: Optional[Union[str,
                                                               int]] = None):
            if isinstance(loaded_shard_id, str):
                _loaded_shard_id = ["q", "k", "v"].index(loaded_shard_id)
            else:
                _loaded_shard_id = loaded_shard_id or 0

            shard_sizes = [i[1] - i[0] for i in indice_sizes]
            offset, end = indice_sizes[_loaded_shard_id]
            param_data = param.data
            if loaded_shard_id is not None:
                param_data = param_data.narrow(
                    narrow_dim,
                    sum(shard_sizes[:_loaded_shard_id]),
                    shard_sizes[_loaded_shard_id],
                )

            # split for TP
            loaded_weight = loaded_weight.narrow(narrow_dim, offset,
                                                 end - offset)
            assert param_data.shape == loaded_weight.shape
            param_data.copy_(loaded_weight)

        return wrap_weight_loader

    def create_weights(self, layer: torch.nn.Module,
                       input_size_per_partition: int,
                       output_partition_sizes: List[int], input_size: int,
                       output_size: int, params_dtype: torch.dtype,
                       **extra_weight_attrs):
        row_parallel_tp_size = input_size // input_size_per_partition
        col_parallel_tp_size = output_size // sum(output_partition_sizes)

        if params_dtype != torch.half and params_dtype != torch.bfloat16:
            raise ValueError(
                "Only half and bfloat16 are currently supported by vptq")
        quant_config = self.quant_config.get("q_proj", self.quant_config)
        quant_config = quant_config.get("gate_proj", quant_config)

        num_codebooks = quant_config["group_num"]
        num_centroids = quant_config["num_centroids"][1]
        group_size = quant_config["group_size"]
        vector_len = quant_config["vector_lens"][1]
        num_res_centroids = quant_config["num_res_centroids"][1]
        enable_residual = num_res_centroids > 0
        enable_norm = quant_config["enable_norm"]
        enable_perm = quant_config["enable_perm"]
        assert not enable_perm, (
            "perm is not absorbed in this model, please process it \
by `pip install vptq && python -m vptq.tools.pre_process \
--input_path xx --output_path xx`")
        assert input_size == group_size
        group_size = input_size_per_partition
        metadata = MetaData()
        metadata.num_centroids = num_centroids
        metadata.num_res_centroids = num_res_centroids
        metadata.vector_len = vector_len
        metadata.group_size = group_size
        layer.metadata = metadata

        num_linears = len(output_partition_sizes)
        orig_weight_loader = extra_weight_attrs['weight_loader']

        if enable_norm:
            wrapped_weight_loader = VPTQLinearMethod.quantized_weight_loader(
                [[(
                    input_size_per_partition * tp_ind,
                    input_size_per_partition * (tp_ind + 1),
                ) for num in output_partition_sizes]
                 for tp_ind in range(row_parallel_tp_size)
                 ][get_tensor_model_parallel_rank() % row_parallel_tp_size],
                0,
            )
            extra_weight_attrs["weight_loader"] = wrapped_weight_loader

            extra_weight_attrs["output_dim"] = 0
            weight_scale = Parameter(torch.empty(input_size_per_partition *
                                                 num_linears,
                                                 dtype=params_dtype),
                                     requires_grad=False)
            weight_bias = Parameter(torch.empty(input_size_per_partition *
                                                num_linears,
                                                dtype=params_dtype),
                                    requires_grad=False)
            set_weight_attrs(weight_scale, extra_weight_attrs)
            set_weight_attrs(weight_bias, extra_weight_attrs)
            layer.register_parameter("weight_scale", weight_scale)
            layer.register_parameter("weight_bias", weight_bias)
            extra_weight_attrs["weight_loader"] = orig_weight_loader

        index_bits = int(math.log2(num_centroids))
        res_index_bits = int(
            math.log2(num_res_centroids)) if enable_residual else 0
        total_index_bits = index_bits + res_index_bits
        packed_groupsize = math.ceil(group_size * total_index_bits / 32)

        indice_sizes = [[(math.floor(num * tp_ind / vector_len),
                          math.ceil(num * (tp_ind + 1) / vector_len))
                         for num in output_partition_sizes]
                        for tp_ind in range(col_parallel_tp_size)]
        tp_output_offset = [[(num * tp_ind) % vector_len
                             for num in output_partition_sizes]
                            for tp_ind in range(col_parallel_tp_size)]
        if col_parallel_tp_size > 1:
            this_rank_indice_sizes = indice_sizes[
                get_tensor_model_parallel_rank()]
        else:
            this_rank_indice_sizes = indice_sizes[0]
        shard_sizes = [i[1] - i[0] for i in this_rank_indice_sizes]
        num_indices = sum(shard_sizes)
        indices = Parameter(torch.empty(
            (num_codebooks, num_indices, packed_groupsize), dtype=torch.int32),
                            requires_grad=False)
        if row_parallel_tp_size == 1:
            wrapped_weight_loader = VPTQLinearMethod.quantized_weight_loader(
                this_rank_indice_sizes)
            extra_weight_attrs['weight_loader'] = wrapped_weight_loader

        set_weight_attrs(
            indices,
            {
                # metadata indicates fixed size concatenated along dim 0
                "output_partition_sizes": output_partition_sizes,
                "output_offset": tp_output_offset,
                "shard_sizes": shard_sizes,
                "input_dim": -1,
            },
        )

        extra_weight_attrs["output_dim"] = 1
        set_weight_attrs(indices, extra_weight_attrs)
        layer.register_parameter("indices", indices)
        extra_weight_attrs['weight_loader'] = orig_weight_loader

        extra_weight_attrs.pop("output_dim")
        extra_weight_attrs["is_metadata"] = True
        centroids = torch.nn.Embedding(num_codebooks * num_linears,
                                       num_centroids * vector_len,
                                       dtype=params_dtype)
        set_weight_attrs(centroids.weight, extra_weight_attrs)
        set_weight_attrs(
            centroids.weight,
            {
                # metadata indicates fixed size concatenated along dim 0
                "codebook_sizes":
                [num_centroids * vector_len for _ in output_partition_sizes],
            },
        )
        layer.centroids = centroids
        # layer.register_parameter("centroids", centroids)
        if enable_residual:
            res_centroids = torch.nn.Embedding(num_codebooks * num_linears,
                                               num_res_centroids * vector_len,
                                               dtype=params_dtype)
            set_weight_attrs(res_centroids.weight, extra_weight_attrs)
            # layer.register_parameter("res_centroids", res_centroids)
            layer.res_centroids = res_centroids
            set_weight_attrs(
                res_centroids.weight,
                {
                    # metadata indicates fixed size concatenated along dim 1
                    "codebook_sizes": [
                        num_res_centroids * vector_len
                        for _ in output_partition_sizes
                    ],
                },
            )

    def apply(
        self,
        layer: torch.nn.Module,
        x: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        weight_scale = layer.weight_scale if hasattr(layer,
                                                     "weight_scale") else None
        weight_bias = layer.weight_bias if hasattr(layer,
                                                   "weight_bias") else None
        perm = layer.perm if hasattr(layer, "perm") else None
        indices = layer.indices
        output_partition_sizes = getattr(indices, "output_partition_sizes", [])
        centroids = layer.centroids.weight
        res_centroids = layer.res_centroids.weight if hasattr(
            layer, "res_centroids") else None

        # fall back all unoptimized formats
        return merged_dequantize_gemm(x, indices, centroids, res_centroids,
                                      weight_scale, weight_bias, perm,
                                      output_partition_sizes, bias,
                                      layer.metadata)



from typing import Any, Dict, Tuple

import pytest
import torch
from PIL.Image import Image
from transformers import AutoTokenizer

from aphrodite.inputs import InputContext, token_inputs
from aphrodite.multimodal import MultiModalRegistry

from ....conftest import _ImageAssets
from ...utils import build_model_context

MODEL = "Qwen/Qwen2-VL-2B-Instruct"
MIN_PIXELS = "min_pixels"
MAX_PIXELS = "max_pixels"


# Fixtures lazy import to avoid initializing CUDA during test collection
# NOTE: Qwen2vl supports multiple input modalities, so it registers multiple
# input mappers.
@pytest.fixture()
def image_input_mapper_for_qwen2_vl():
    from aphrodite.modeling.models.qwen2_vl import (
        image_input_mapper_for_qwen2_vl)
    return image_input_mapper_for_qwen2_vl


@pytest.fixture()
def input_processor_for_qwen2_vl():
    from aphrodite.modeling.models.qwen2_vl import input_processor_for_qwen2_vl
    return input_processor_for_qwen2_vl


@pytest.fixture()
def qwen2_vl_context() -> InputContext:
    return build_model_context(model_name=MODEL)


@pytest.fixture()
def get_max_qwen2_vl_image_tokens():
    from aphrodite.modeling.models.qwen2_vl import (
        get_max_qwen2_vl_image_tokens)
    return get_max_qwen2_vl_image_tokens


@pytest.fixture()
def dummy_data_for_qwen2_vl():
    from aphrodite.modeling.models.qwen2_vl import dummy_data_for_qwen2_vl
    return dummy_data_for_qwen2_vl


@pytest.mark.parametrize("mm_processor_kwargs,expected_max_tokens", [
    ({}, 1225),
    ({
        MIN_PIXELS: 64**2,
        MAX_PIXELS: 512**2
    }, 324),
])
def test_qwen2_vl_max_image_tokens(get_max_qwen2_vl_image_tokens,
                                   qwen2_vl_context: InputContext,
                                   mm_processor_kwargs: Dict[str, Any],
                                   expected_max_tokens: int):
    """Ensure that the max token calc handles min/max pixels properly."""
    actual_max_tokens = get_max_qwen2_vl_image_tokens(qwen2_vl_context,
                                                      **mm_processor_kwargs)
    assert actual_max_tokens == expected_max_tokens


@pytest.mark.parametrize("mm_processor_kwargs,token_count,img_size", [
    [{}, 1225, (980, 980)],
    [{
        MIN_PIXELS: 64**2,
        MAX_PIXELS: 512**2
    }, 324, (504, 504)],
])
def test_qwen2_vl_dummy_data(dummy_data_for_qwen2_vl,
                             qwen2_vl_context: InputContext,
                             mm_processor_kwargs: Dict[str, Any],
                             token_count: int, img_size: Tuple[int, int]):
    """Ensure that the dummy data handles min/max pixels properly."""
    seq_len = 3000
    hf_config = qwen2_vl_context.get_hf_config()
    image_token_id = hf_config.image_token_id

    # NOTE: video value is required, but isn't actually used
    # when making the dummy data except for error handling currently
    seq_data, mm_data = dummy_data_for_qwen2_vl(qwen2_vl_context, seq_len, {
        "image": 1,
        "video": 0
    }, **mm_processor_kwargs)

    # Ensure we have the right number of placeholders for min/max pixel values
    assert seq_data.get_token_ids().count(image_token_id) == token_count

    # Ensure the images were resized correctly
    image = mm_data["image"]
    assert isinstance(image, Image)
    assert image.size == img_size


@pytest.mark.parametrize("mm_processor_kwargs,num_placeholders", [
    ({}, 1426),
    ({
        MIN_PIXELS: 64**2,
        MAX_PIXELS: 512**2
    }, 330),
])
def test_input_processor(input_processor_for_qwen2_vl,
                         qwen2_vl_context: InputContext,
                         image_assets: _ImageAssets, num_placeholders: int,
                         mm_processor_kwargs: Dict[str, Any]):
    """Ensure that the image processor handles min/max pixels properly."""
    tokenizer = AutoTokenizer.from_pretrained(MODEL)
    prompt = "<|vision_start|><|image_pad|><|vision_end|>"

    image = image_assets[0].pil_image
    hf_config = qwen2_vl_context.get_hf_config()
    image_token_id = hf_config.image_token_id

    inputs = token_inputs(prompt_token_ids=tokenizer.encode(prompt),
                          prompt=prompt,
                          multi_modal_data={"image": [image]})

    processed_inputs = input_processor_for_qwen2_vl(qwen2_vl_context, inputs,
                                                    **mm_processor_kwargs)
    assert processed_inputs["prompt_token_ids"].count(
        image_token_id) == num_placeholders
    assert len(processed_inputs["multi_modal_data"]["image"]) == 1


@pytest.mark.parametrize("mm_processor_kwargs,pixels_shape", [
    ({}, [5704, 1176]),
    ({
        MIN_PIXELS: 64**2,
        MAX_PIXELS: 512**2
    }, [1320, 1176]),
])
def test_image_mapper_override(qwen2_vl_context: InputContext,
                               image_assets: _ImageAssets,
                               mm_processor_kwargs: Dict[str, Any],
                               pixels_shape: Tuple[int, int]):
    """Ensure that the image mapper handles min/max pixels properly."""
    mm_registry = MultiModalRegistry()
    mm_registry.init_mm_limits_per_prompt(qwen2_vl_context.model_config)

    image = image_assets[0].pil_image

    mapped_output = mm_registry.map_input(
        qwen2_vl_context.model_config,
        {"image": image},
        mm_processor_kwargs=mm_processor_kwargs,
    )

    # Dimension 0 of pixel values should match the product of image_grid_thw
    actual_pixels_shape = mapped_output["pixel_values"].shape
    assert list(actual_pixels_shape) == pixels_shape
    assert actual_pixels_shape[0] == torch.prod(
        mapped_output["image_grid_thw"])



import os
from typing import Optional, Union

from loguru import logger
from transformers.dynamic_module_utils import get_class_from_dynamic_module

import aphrodite.common.envs as envs


def try_get_class_from_dynamic_module(
    class_reference: str,
    pretrained_model_name_or_path: str,
    cache_dir: Optional[Union[str, os.PathLike]] = None,
    force_download: bool = False,
    resume_download: Optional[bool] = None,
    proxies: Optional[dict[str, str]] = None,
    token: Optional[Union[bool, str]] = None,
    revision: Optional[str] = None,
    local_files_only: bool = False,
    repo_type: Optional[str] = None,
    code_revision: Optional[str] = None,
    warn_on_fail: bool = True,
    **kwargs,
) -> Optional[type]:
    """
    As [transformers.dynamic_module_utils.get_class_from_dynamic_module][],
    but ignoring any errors.
    """
    try:
        return get_class_from_dynamic_module(
            class_reference,
            pretrained_model_name_or_path,
            cache_dir=cache_dir,
            force_download=force_download,
            resume_download=resume_download,
            proxies=proxies,
            token=token,
            revision=revision,
            local_files_only=local_files_only,
            repo_type=repo_type,
            code_revision=code_revision,
            **kwargs,
        )
    except Exception:
        location = "ModelScope" if envs.APHRODITE_USE_MODELSCOPE else "HF Hub"

        if warn_on_fail:
            logger.warning(
                "Unable to load {} from {} on {}.",
                class_reference,
                pretrained_model_name_or_path,
                location,
                exc_info=True,
            )

        return None



from typing import Optional

from loguru import logger

from aphrodite.common import envs
from aphrodite.common.logger import log_once
from aphrodite.platforms import current_platform

if current_platform.is_cuda():
    from aphrodite import _custom_ops as ops
    reshape_and_cache_flash = ops.reshape_and_cache_flash
    from aphrodite.aphrodite_flash_attn import (flash_attn_varlen_func,
                                                get_scheduler_metadata)
elif current_platform.is_xpu():
    from aphrodite._ipex_ops import ipex_ops as ops
    reshape_and_cache_flash = ops.reshape_and_cache_flash
    flash_attn_varlen_func = ops.flash_attn_varlen_func
    get_scheduler_metadata = ops.get_scheduler_metadata


def get_flash_attn_version(requires_alibi: bool = False) -> Optional[int]:
    # import here to avoid circular dependencies
    from aphrodite.platforms import current_platform
    if current_platform.is_xpu():
        return 2
    try:
        from aphrodite.aphrodite_flash_attn.flash_attn_interface import (
            fa_version_unsupported_reason, is_fa_version_supported)
        device_capability = current_platform.get_device_capability()

        assert device_capability is not None

        # 1. default version depending on platform
        fa_version = 3 if (device_capability.major == 9
                           and is_fa_version_supported(3)) else 2

        # 2. override if passed by environment
        if envs.APHRODITE_FLASH_ATTN_VERSION is not None:
            assert envs.APHRODITE_FLASH_ATTN_VERSION in [2, 3]
            fa_version = envs.APHRODITE_FLASH_ATTN_VERSION

        # 3. fallback for unsupported combinations
        if device_capability.major == 10 and fa_version == 3:
            log_once(
                "WARNING",
                "Cannot use FA version 3 on Blackwell platform "
                "defaulting to FA version 2.")
            fa_version = 2

        if requires_alibi and fa_version == 3:
            log_once(
                "WARNING",
                "Cannot use FA version 3 with ALiBi, "
                "defaulting to FA version 2.")
            fa_version = 2

        if not is_fa_version_supported(fa_version):
            logger.error("Cannot use FA version {} is not supported due to {}",
                         fa_version, fa_version_unsupported_reason(fa_version))

        assert is_fa_version_supported(fa_version)
        return fa_version
    except (ImportError, AssertionError):
        return None


def flash_attn_supports_fp8() -> bool:
    return get_flash_attn_version() == 3 and \
        current_platform.get_device_capability().major == 9


def is_flash_attn_varlen_func_available() -> bool:
    return current_platform.is_cuda() or current_platform.is_xpu()



import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from aphrodite.common.logger import log_once

if TYPE_CHECKING:
    # Avoid circular import.
    from aphrodite.endpoints.context import ConversationContext


def validate_gpt_oss_install():
    """
    Check if the gpt-oss is installed and its version is at least 0.0.3.
    If not, raise an ImportError.
    """
    from importlib.metadata import PackageNotFoundError, version

    from packaging.version import InvalidVersion, Version

    try:
        pkg_version_str = version("gpt_oss")  # e.g., "0.0.5"
        pkg_version = Version(pkg_version_str)
    except PackageNotFoundError:
        raise ImportError("Package 'gpt_oss' is not installed.") from None
    except InvalidVersion as e:
        raise ImportError(
            f"Invalid version string for 'gpt_oss': {e}") from None

    if pkg_version < Version("0.0.3"):
        raise ImportError(
            f"gpt_oss >= 0.0.3 is required, but {pkg_version} is installed."
        ) from None


class Tool(ABC):

    @abstractmethod
    async def get_result(self, context: "ConversationContext") -> Any:
        pass


class HarmonyBrowserTool(Tool):

    def __init__(self):
        self.enabled = True
        exa_api_key = os.getenv("EXA_API_KEY")
        if not exa_api_key:
            self.enabled = False
            log_once(
                "WARNING",
                "EXA_API_KEY is not set, browsing is disabled")
            return

        try:
            validate_gpt_oss_install()
            from gpt_oss.tools.simple_browser import SimpleBrowserTool
            from gpt_oss.tools.simple_browser.backend import ExaBackend
        except ImportError as e:
            self.enabled = False
            log_once(
                "WARNING",
                "gpt_oss is not installed properly ({}), browsing is disabled",
                e)
            return

        browser_backend = ExaBackend(source="web", api_key=exa_api_key)
        self.browser_tool = SimpleBrowserTool(backend=browser_backend)
        log_once(
            "INFO",
            "Browser tool initialized")

    async def get_result(self, context: "ConversationContext") -> Any:
        from aphrodite.endpoints.context import HarmonyContext
        assert isinstance(context, HarmonyContext)
        last_msg = context.messages[-1]
        tool_output_msgs = []
        async for msg in self.browser_tool.process(last_msg):
            tool_output_msgs.append(msg)
        return tool_output_msgs

    @property
    def tool_config(self) -> Any:
        return self.browser_tool.tool_config


class HarmonyPythonTool(Tool):

    def __init__(self):
        self.enabled = True

        try:
            validate_gpt_oss_install()
            from gpt_oss.tools.python_docker.docker_tool import PythonTool
        except ImportError as e:
            self.enabled = False
            log_once(
                "WARNING",
                "gpt_oss is not installed properly ({}), code interpreter is "
                "disabled",
                e)
            return

        self.python_tool = PythonTool()
        log_once(
            "INFO",
            "Code interpreter tool initialized")

    async def get_result(self, context: "ConversationContext") -> Any:
        from aphrodite.endpoints.context import HarmonyContext
        assert isinstance(context, HarmonyContext)
        last_msg = context.messages[-1]
        tool_output_msgs = []
        async for msg in self.python_tool.process(last_msg):
            tool_output_msgs.append(msg)
        return tool_output_msgs

    @property
    def tool_config(self) -> Any:
        return self.python_tool.tool_config



"""
This example shows how to use Aphrodite for running offline inference with
the correct prompt format on vision language models for multimodal embedding.

For most models, the prompt format should follow corresponding examples
on HuggingFace model repository.
"""
from argparse import Namespace
from typing import Literal, NamedTuple, Optional, TypedDict, Union, get_args

from PIL.Image import Image

from aphrodite import LLM
from aphrodite.utils import FlexibleArgumentParser
from aphrodite.multimodal.utils import fetch_image


class TextQuery(TypedDict):
    modality: Literal["text"]
    text: str


class ImageQuery(TypedDict):
    modality: Literal["image"]
    image: Image


class TextImageQuery(TypedDict):
    modality: Literal["text+image"]
    text: str
    image: Image


QueryModality = Literal["text", "image", "text+image"]
Query = Union[TextQuery, ImageQuery, TextImageQuery]


class ModelRequestData(NamedTuple):
    llm: LLM
    prompt: str
    image: Optional[Image]


def run_e5_v(query: Query):
    llama3_template = '<|start_header_id|>user<|end_header_id|>\n\n{}<|eot_id|><|start_header_id|>assistant<|end_header_id|>\n\n \n'  # noqa: E501

    if query["modality"] == "text":
        text = query["text"]
        prompt = llama3_template.format(
            f"{text}\nSummary above sentence in one word: ")
        image = None
    elif query["modality"] == "image":
        prompt = llama3_template.format(
            "<image>\nSummary above image in one word: ")
        image = query["image"]
    else:
        modality = query['modality']
        raise ValueError(f"Unsupported query modality: '{modality}'")

    llm = LLM(
        model="royokong/e5-v",
        task="embedding",
        max_model_len=4096,
    )

    return ModelRequestData(
        llm=llm,
        prompt=prompt,
        image=image,
    )


def run_vlm2vec(query: Query):
    if query["modality"] == "text":
        text = query["text"]
        prompt = f"Find me an everyday image that matches the given caption: {text}"  # noqa: E501
        image = None
    elif query["modality"] == "image":
        prompt = "<|image_1|> Find a day-to-day image that looks similar to the provided image."  # noqa: E501
        image = query["image"]
    elif query["modality"] == "text+image":
        text = query["text"]
        prompt = f"<|image_1|> Represent the given image with the following question: {text}"  # noqa: E501
        image = query["image"]
    else:
        modality = query['modality']
        raise ValueError(f"Unsupported query modality: '{modality}'")

    llm = LLM(
        model="TIGER-Lab/VLM2Vec-Full",
        task="embedding",
        trust_remote_code=True,
        mm_processor_kwargs={"num_crops": 4},
    )

    return ModelRequestData(
        llm=llm,
        prompt=prompt,
        image=image,
    )


def get_query(modality: QueryModality):
    if modality == "text":
        return TextQuery(modality="text", text="A dog sitting in the grass")

    if modality == "image":
        return ImageQuery(
            modality="image",
            image=fetch_image(
                "https://upload.wikimedia.org/wikipedia/commons/thumb/4/47/American_Eskimo_Dog.jpg/360px-American_Eskimo_Dog.jpg"  # noqa: E501
            ),
        )

    if modality == "text+image":
        return TextImageQuery(
            modality="text+image",
            text="A cat standing in the snow.",
            image=fetch_image(
                "https://upload.wikimedia.org/wikipedia/commons/thumb/b/b6/Felis_catus-cat_on_snow.jpg/179px-Felis_catus-cat_on_snow.jpg"  # noqa: E501
            ),
        )

    msg = f"Modality {modality} is not supported."
    raise ValueError(msg)


def run_encode(model: str, modality: QueryModality):
    query = get_query(modality)
    req_data = model_example_map[model](query)

    mm_data = {}
    if req_data.image is not None:
        mm_data["image"] = req_data.image

    outputs = req_data.llm.encode({
        "prompt": req_data.prompt,
        "multi_modal_data": mm_data,
    })

    for output in outputs:
        print(output.outputs.embedding)


def main(args: Namespace):
    run_encode(args.model_name, args.modality)


model_example_map = {
    "e5_v": run_e5_v,
    "vlm2vec": run_vlm2vec,
}

if __name__ == "__main__":
    parser = FlexibleArgumentParser(
        description='Demo on using Aphrodite for offline inference with '
        'vision language models for multimodal embedding')
    parser.add_argument('--model-name',
                        '-m',
                        type=str,
                        default="vlm2vec",
                        choices=model_example_map.keys(),
                        help='The name of the embedding model.')
    parser.add_argument('--modality',
                        type=str,
                        default="image",
                        choices=get_args(QueryModality),
                        help='Modality of the input.')
    args = parser.parse_args()
    main(args)



# Copyright 2024 Cohere and the HuggingFace Inc. team. All rights reserved.
#
# This code is based on EleutherAI's GPT-NeoX library and the GPT-NeoX
# and OPT implementations in this library. It has been modified from its
# original forms to accommodate minor architectural differences compared
# to GPT-NeoX and OPT used by the Meta AI team that trained the model.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file is based on the LLama model definition file in transformers
"""PyTorch Cohere model."""
from collections.abc import Iterable
from typing import Optional, Union

import torch
from torch import nn
from transformers import Cohere2Config, CohereConfig

from aphrodite.attention import Attention
from aphrodite.config import AphroditeConfig, CacheConfig
from aphrodite.common.sequence import IntermediateTensors
from aphrodite.compilation.decorators import support_torch_compile
from aphrodite.distributed import (get_pp_group,
                                   get_tensor_model_parallel_world_size)
from aphrodite.modeling.layers.activation import SiluAndMul
from aphrodite.modeling.layers.linear import (MergedColumnParallelLinear,
                                              QKVParallelLinear,
                                              RowParallelLinear)
from aphrodite.modeling.layers.logits_processor import LogitsProcessor
from aphrodite.modeling.layers.rotary_embedding import get_rope
from aphrodite.modeling.layers.vocab_parallel_embedding import (
    VocabParallelEmbedding)
from aphrodite.modeling.model_loader.weight_utils import (
    default_weight_loader, maybe_remap_kv_scale_name,
    row_parallel_weight_loader)
from aphrodite.modeling.sampling_metadata import SamplingMetadata
from aphrodite.modeling.utils import set_weight_attrs
from aphrodite.platforms import current_platform
from aphrodite.quantization import QuantizationConfig

from .interfaces import SupportsLoRA, SupportsPP, SupportsQuant
from .utils import (AutoWeightsLoader, extract_layer_index,
                    is_pp_missing_parameter,
                    make_empty_intermediate_tensors_factory, make_layers,
                    maybe_prefix)


@torch.compile(backend=current_platform.simple_compile_backend)
def layer_norm_func(hidden_states, weight, variance_epsilon):
    input_dtype = hidden_states.dtype
    hidden_states = hidden_states.to(torch.float32)
    mean = hidden_states.mean(-1, keepdim=True)
    variance = (hidden_states - mean).pow(2).mean(-1, keepdim=True)
    hidden_states = (hidden_states - mean) * torch.rsqrt(variance +
                                                         variance_epsilon)
    hidden_states = weight.to(torch.float32) * hidden_states
    return hidden_states.to(input_dtype)


class LayerNorm(nn.Module):

    def __init__(self, param_shape=None, eps=1e-5):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(param_shape))
        self.variance_epsilon = eps
        set_weight_attrs(self.weight,
                         {"weight_loader": row_parallel_weight_loader})

    def forward(self, hidden_states, residuals=None):
        hidden_states = layer_norm_func(hidden_states, self.weight,
                                        self.variance_epsilon)
        return hidden_states, residuals


# Copied from transformers.models.llama.modeling_llama.LlamaMLP Llama->Cohere
class CohereMLP(nn.Module):

    def __init__(
        self,
        config: Union[CohereConfig, Cohere2Config],
        quant_config: Optional[QuantizationConfig] = None,
        prefix: str = "",
    ):
        super().__init__()
        self.config = config
        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size
        self.gate_up_proj = MergedColumnParallelLinear(
            self.hidden_size,
            [self.intermediate_size] * 2,
            bias=False,
            quant_config=quant_config,
            prefix=f"{prefix}.gate_up_proj",
        )
        self.down_proj = RowParallelLinear(
            self.intermediate_size,
            self.hidden_size,
            bias=False,
            quant_config=quant_config,
            prefix=f"{prefix}.down_proj",
        )
        self.act_fn = SiluAndMul()

    def forward(self, x):
        gate_up, _ = self.gate_up_proj(x)
        x = self.act_fn(gate_up)
        x, _ = self.down_proj(x)
        return x


class CohereAttention(nn.Module):

    def __init__(
        self,
        config: Union[CohereConfig, Cohere2Config],
        cache_config: Optional[CacheConfig] = None,
        quant_config: Optional[QuantizationConfig] = None,
        prefix: str = "",
    ):
        super().__init__()
        tp_size = get_tensor_model_parallel_world_size()
        self.config = config
        self.attention_dropout = config.attention_dropout
        self.hidden_size = config.hidden_size
        self.total_num_heads = config.num_attention_heads
        self.num_heads = self.total_num_heads // tp_size
        self.head_dim = self.hidden_size // self.total_num_heads
        self.total_num_kv_heads = config.num_key_value_heads
        if self.total_num_kv_heads >= tp_size:
            # Number of KV heads is greater than TP size, so we partition
            # the KV heads across multiple tensor parallel GPUs.
            assert self.total_num_kv_heads % tp_size == 0
        else:
            # Number of KV heads is less than TP size, so we replicate
            # the KV heads across multiple tensor parallel GPUs.
            assert tp_size % self.total_num_kv_heads == 0
        self.num_kv_heads = max(1, self.total_num_kv_heads // tp_size)
        self.q_size = self.num_heads * self.head_dim
        self.kv_size = self.num_kv_heads * self.head_dim
        self.scaling = self.head_dim**-0.5
        self.max_position_embeddings = getattr(
            config, "model_max_length", None) or getattr(
                config, "max_position_embeddings", 8192)
        self.rope_theta = config.rope_theta
        self.rope_scaling = getattr(config, "rope_scaling", None)
        self.use_qk_norm = getattr(config, "use_qk_norm", False)
        self.qkv_proj = QKVParallelLinear(
            self.hidden_size,
            self.head_dim,
            self.total_num_heads,
            self.total_num_kv_heads,
            bias=False,
            quant_config=quant_config,
            prefix=f"{prefix}.qkv_proj",
        )
        self.o_proj = RowParallelLinear(
            self.total_num_heads * self.head_dim,
            self.hidden_size,
            bias=False,
            quant_config=quant_config,
            prefix=f"{prefix}.o_proj",
        )
        self.rotary_emb = get_rope(
            self.head_dim,
            rotary_dim=self.head_dim,
            max_position=self.max_position_embeddings,
            base=self.rope_theta,
            rope_scaling=self.rope_scaling,
            is_neox_style=False,
        )

        # Model v2 has interleaved sliding windows, v1 does not
        self.v1 = isinstance(config, CohereConfig)

        self.sliding_window = None
        if not self.v1:
            layer_idx = extract_layer_index(prefix)
            if config.layer_types[layer_idx] == "sliding_attention":
                self.sliding_window = config.sliding_window

        self.attn = Attention(self.num_heads,
                              self.head_dim,
                              self.scaling,
                              num_kv_heads=self.num_kv_heads,
                              cache_config=cache_config,
                              quant_config=quant_config,
                              per_layer_sliding_window=self.sliding_window,
                              prefix=f"{prefix}.attn")
        if self.use_qk_norm:
            self.q_norm = LayerNorm(param_shape=(self.num_heads,
                                                 self.head_dim),
                                    eps=config.layer_norm_eps)
            self.k_norm = LayerNorm(param_shape=(self.num_kv_heads,
                                                 self.head_dim),
                                    eps=config.layer_norm_eps)

    def _apply_qk_norm(self, q, k):
        q = q.view(*q.shape[:-1], -1, self.head_dim)
        k = k.view(*k.shape[:-1], -1, self.head_dim)
        q, _ = self.q_norm(q)
        k, _ = self.k_norm(k)
        q = q.view(*q.shape[:-2], -1)
        k = k.view(*k.shape[:-2], -1)
        return q, k

    def forward(
        self,
        positions: torch.Tensor,
        hidden_states: torch.Tensor,
    ) -> torch.Tensor:
        qkv, _ = self.qkv_proj(hidden_states)
        q, k, v = qkv.split([self.q_size, self.kv_size, self.kv_size], dim=-1)
        if self.use_qk_norm:
            q, k = self._apply_qk_norm(q, k)
        if self.v1 or self.sliding_window:
            q, k = self.rotary_emb(positions, q, k)
        attn_output = self.attn(q, k, v)
        output, _ = self.o_proj(attn_output)
        return output


class CohereDecoderLayer(nn.Module):

    def __init__(self,
                 config: Union[CohereConfig, Cohere2Config],
                 cache_config: Optional[CacheConfig] = None,
                 quant_config: Optional[QuantizationConfig] = None,
                 prefix: str = ""):
        super().__init__()
        self.hidden_size = config.hidden_size

        self.self_attn = CohereAttention(config,
                                         cache_config,
                                         quant_config=quant_config,
                                         prefix=f"{prefix}.self_attn")

        self.mlp = CohereMLP(config,
                             quant_config=quant_config,
                             prefix=f"{prefix}.mlp")
        self.input_layernorm = LayerNorm(param_shape=(config.hidden_size),
                                         eps=config.layer_norm_eps)

    def forward(
        self,
        positions: torch.Tensor,
        hidden_states: torch.Tensor,
        residual: Optional[torch.Tensor],
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Self Attention
        residual = hidden_states
        hidden_states, residual = self.input_layernorm(hidden_states, residual)
        hidden_states_attention = self.self_attn(
            positions=positions,
            hidden_states=hidden_states,
        )
        hidden_states_mlp = self.mlp(hidden_states)
        # Add everything together
        hidden_states = residual + hidden_states_attention + hidden_states_mlp

        return hidden_states, residual


@support_torch_compile
class CohereModel(nn.Module):

    def __init__(self, *, aphrodite_config: AphroditeConfig, prefix: str = ""):
        super().__init__()

        config = aphrodite_config.model_config.hf_config
        cache_config = aphrodite_config.cache_config
        quant_config = aphrodite_config.quant_config
        lora_config = aphrodite_config.lora_config
        self.quant_config = quant_config

        self.config = config
        lora_vocab = (lora_config.lora_extra_vocab_size *
                      (lora_config.max_loras or 1)) if lora_config else 0
        self.vocab_size = config.vocab_size + lora_vocab
        self.org_vocab_size = config.vocab_size
        self.embed_tokens = VocabParallelEmbedding(config.vocab_size,
                                                   config.hidden_size)
        self.start_layer, self.end_layer, self.layers = make_layers(
            config.num_hidden_layers,
            lambda prefix: CohereDecoderLayer(
                config, cache_config, quant_config, prefix=prefix),
            prefix=f"{prefix}.layers")
        self.norm = LayerNorm(param_shape=(config.hidden_size),
                              eps=config.layer_norm_eps)
        self.make_empty_intermediate_tensors = (
            make_empty_intermediate_tensors_factory(
                ["hidden_states", "residual"], config.hidden_size))

    def get_input_embeddings(self, input_ids: torch.Tensor) -> torch.Tensor:
        return self.embed_tokens(input_ids)

    def forward(
        self,
        input_ids: torch.Tensor,
        positions: torch.Tensor,
        intermediate_tensors: Optional[IntermediateTensors],
        inputs_embeds: Optional[torch.Tensor] = None,
    ) -> Union[torch.Tensor, IntermediateTensors]:
        if get_pp_group().is_first_rank:
            if inputs_embeds is not None:
                hidden_states = inputs_embeds
            else:
                hidden_states = self.get_input_embeddings(input_ids)
            residual = None
        else:
            assert intermediate_tensors is not None
            hidden_states = intermediate_tensors["hidden_states"]
            residual = intermediate_tensors["residual"]
        for layer in self.layers[self.start_layer:self.end_layer]:
            hidden_states, residual = layer(
                positions,
                hidden_states,
                residual,
            )
        if not get_pp_group().is_last_rank:
            return IntermediateTensors({
                "hidden_states": hidden_states,
                "residual": residual
            })
        hidden_states, _ = self.norm(hidden_states, residual)
        return hidden_states

    def load_weights(self, weights: Iterable[tuple[str,
                                                   torch.Tensor]]) -> set[str]:
        stacked_params_mapping = [
            # (param_name, shard_name, shard_id)
            ("qkv_proj", "q_proj", "q"),
            ("qkv_proj", "k_proj", "k"),
            ("qkv_proj", "v_proj", "v"),
            ("gate_up_proj", "gate_proj", 0),
            ("gate_up_proj", "up_proj", 1),
        ]
        params_dict = dict(self.named_parameters())
        loaded_params: set[str] = set()
        for name, loaded_weight in weights:
            if (self.quant_config is not None and
                (scale_name := self.quant_config.get_cache_scale(name))):
                # Loading kv cache quantization scales
                param = params_dict[scale_name]
                weight_loader = getattr(param, "weight_loader",
                                        default_weight_loader)
                loaded_weight = (loaded_weight if loaded_weight.dim() == 0 else
                                 loaded_weight[0])
                weight_loader(param, loaded_weight)
                loaded_params.add(scale_name)
                continue

            for param_name, shard_name, shard_id in stacked_params_mapping:
                if shard_name not in name:
                    continue
                name = name.replace(shard_name, param_name)
                # Skip loading extra bias for GPTQ models.
                if name.endswith(".bias") and name not in params_dict:
                    continue
                if is_pp_missing_parameter(name, self):
                    continue
                param = params_dict[name]
                weight_loader = param.weight_loader
                weight_loader(param, loaded_weight, shard_id)
                break
            else:
                # Skip loading extra bias for GPTQ models.
                if name.endswith(".bias") and name not in params_dict:
                    continue
                # Remapping the name of FP8 kv-scale.
                name = maybe_remap_kv_scale_name(name, params_dict)
                if name is None:
                    continue

                if is_pp_missing_parameter(name, self):
                    continue
                param = params_dict[name]
                weight_loader = getattr(param, "weight_loader",
                                        default_weight_loader)
                weight_loader(param, loaded_weight)
            loaded_params.add(name)
        return loaded_params


class CohereForCausalLM(nn.Module, SupportsLoRA, SupportsPP, SupportsQuant):
    packed_modules_mapping = {
        "qkv_proj": [
            "q_proj",
            "k_proj",
            "v_proj",
        ],
        "gate_up_proj": [
            "gate_proj",
            "up_proj",
        ],
    }
    # LoRA specific attributes
    embedding_modules = {"embed_tokens": "input_embeddings"}

    def __init__(self, *, aphrodite_config: AphroditeConfig, prefix: str = ""):
        super().__init__()
        config = aphrodite_config.model_config.hf_config
        quant_config = aphrodite_config.quant_config
        lora_config = aphrodite_config.lora_config
        self.config = config
        # currently all existing command R models have `tie_word_embeddings`
        # enabled
        assert config.tie_word_embeddings
        self.unpadded_vocab_size = config.vocab_size
        if lora_config:
            self.unpadded_vocab_size += lora_config.lora_extra_vocab_size
        self.quant_config = quant_config
        self.logits_processor = LogitsProcessor(self.unpadded_vocab_size,
                                                config.vocab_size,
                                                scale=config.logit_scale)
        self.model = CohereModel(aphrodite_config=aphrodite_config,
                                 prefix=maybe_prefix(prefix, "model"))
        self.make_empty_intermediate_tensors = (
            self.model.make_empty_intermediate_tensors)

    def get_input_embeddings(self, input_ids: torch.Tensor) -> torch.Tensor:
        return self.model.get_input_embeddings(input_ids)

    @torch.no_grad()
    def forward(
        self,
        input_ids: torch.Tensor,
        positions: torch.Tensor,
        intermediate_tensors: Optional[IntermediateTensors] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
    ) -> Union[torch.Tensor, IntermediateTensors]:
        hidden_states = self.model(input_ids, positions, intermediate_tensors,
                                   inputs_embeds)
        return hidden_states

    def compute_logits(
        self,
        hidden_states: torch.Tensor,
        sampling_metadata: SamplingMetadata,
    ) -> Optional[torch.Tensor]:
        is_not_lora = hasattr(self.model.embed_tokens, 'weight')
        if is_not_lora:
            logits = self.logits_processor(self.model.embed_tokens,
                                           hidden_states, sampling_metadata)
        else:
            logits = self.logits_processor(self.model.embed_tokens.base_layer,
                                           hidden_states, sampling_metadata)

        return logits

    def load_weights(self, weights: Iterable[tuple[str,
                                                   torch.Tensor]]) -> set[str]:
        loader = AutoWeightsLoader(
            self, skip_prefixes=["lm_head", "rotary_emb.inv_freq"])
        return loader.load_weights(weights)



from typing import Optional, Union

import torch

import aphrodite._custom_ops as ops


def maybe_contiguous(x):
    return x.contiguous() if x is not None and x.stride(-1) != 1 else x


def _flash_attn_forward(
    q, k, v, dropout_p, softmax_scale, causal,
    window_size, softcap, alibi_slopes,
    return_softmax, *, out=None
):
    q, k, v = [maybe_contiguous(x) for x in (q, k, v)]
    (out, q, k, v, out_padded, softmax_lse,
     S_dmask, rng_state) = ops.fwd(
        q=q,
        k=k,
        v=v,
        out=out,
        alibi_slopes=alibi_slopes,
        dropout_p=dropout_p,
        softmax_scale=softmax_scale,
        causal=causal,
        window_size_left=window_size[0],
        window_size_right=window_size[1],
        softcap=softcap,
        return_softmax=return_softmax,
        gen=None,
    )  # type: ignore
    return out, q, k, v, out_padded, softmax_lse, S_dmask, rng_state


def _flash_attn_varlen_forward(
    q,
    k,
    v,
    cu_seqlens_q,
    cu_seqlens_k,
    max_seqlen_q,
    max_seqlen_k,
    dropout_p,
    softmax_scale,
    causal,
    window_size,
    softcap,
    alibi_slopes,
    return_softmax,
    block_table,
    *,
    out=None
):
    q, k, v = [maybe_contiguous(x) for x in (q, k, v)]
    (out, q, k, v, out_padded, softmax_lse,
     S_dmask, rng_state) = ops.varlen_fwd(
        q=q,
        k=k,
        v=v,
        cu_seqlens_q=cu_seqlens_q,
        cu_seqlens_k=cu_seqlens_k,
        max_seqlen_q=max_seqlen_q,
        max_seqlen_k=max_seqlen_k,
        dropout_p=dropout_p,
        softmax_scale=softmax_scale,
        causal=causal,
        window_size_left=window_size[0],
        window_size_right=window_size[1],
        softcap=softcap,
        alibi_slopes=alibi_slopes,
        block_table=block_table,
        return_softmax=return_softmax,
        gen=None,
        out=out,
        seqused_k=None,
        zero_tensors=False,
    )  # type: ignore
    return out, q, k, v, out_padded, softmax_lse, S_dmask, rng_state


class FlashAttnFunc(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        q,
        k,
        v,
        dropout_p,
        softmax_scale,
        causal,
        window_size,
        softcap,
        alibi_slopes,
        deterministic,
        return_softmax,
        out=None,
    ):
        if softmax_scale is None:
            softmax_scale = q.shape[-1] ** (-0.5)
        (out, q, k, v, out_padded, softmax_lse,
         S_dmask, rng_state) = _flash_attn_forward(
            q,
            k,
            v,
            dropout_p,
            softmax_scale,
            causal=causal,
            window_size=window_size,
            softcap=softcap,
            alibi_slopes=alibi_slopes,
            return_softmax=return_softmax and dropout_p > 0,
            out=out,
        )
        ctx.save_for_backward(q, k, v, out_padded, softmax_lse, rng_state)
        ctx.dropout_p = dropout_p
        ctx.softmax_scale = softmax_scale
        ctx.causal = causal
        ctx.window_size = window_size
        ctx.softcap = softcap
        ctx.alibi_slopes = alibi_slopes
        ctx.deterministic = deterministic
        return out if not return_softmax else (out, softmax_lse, S_dmask)


class FlashAttnVarlenFunc(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        q,
        k,
        v,
        cu_seqlens_q,
        cu_seqlens_k,
        max_seqlen_q,
        max_seqlen_k,
        dropout_p,
        softmax_scale,
        causal,
        window_size,
        softcap,
        alibi_slopes,
        deterministic,
        return_softmax,
        block_table,
        out=None,
    ):
        if softmax_scale is None:
            softmax_scale = q.shape[-1] ** (-0.5)
        (out, q, k, v, out_padded, softmax_lse,
         S_dmask, rng_state) = _flash_attn_varlen_forward(
            q,
            k,
            v,
            cu_seqlens_q,
            cu_seqlens_k,
            max_seqlen_q,
            max_seqlen_k,
            dropout_p,
            softmax_scale,
            causal=causal,
            window_size=window_size,
            softcap=softcap,
            alibi_slopes=alibi_slopes,
            return_softmax=return_softmax and dropout_p > 0,
            block_table=block_table,
            out=out,
        )
        ctx.save_for_backward(
            q, k, v, out_padded, softmax_lse, cu_seqlens_q,
            cu_seqlens_k, rng_state
        )
        ctx.dropout_p = dropout_p
        ctx.max_seqlen_q = max_seqlen_q
        ctx.max_seqlen_k = max_seqlen_k
        ctx.softmax_scale = softmax_scale
        ctx.causal = causal
        ctx.window_size = window_size
        ctx.softcap = softcap
        ctx.alibi_slopes = alibi_slopes
        ctx.deterministic = deterministic
        return out if not return_softmax else (out, softmax_lse, S_dmask)


def flash_attn_varlen_func(
    q,
    k,
    v,
    cu_seqlens_q,
    cu_seqlens_k,
    max_seqlen_q,
    max_seqlen_k,
    dropout_p=0.0,
    softmax_scale=None,
    causal=False,
    window_size=(-1, -1),  # -1 means infinite context window
    softcap=0.0, # 0.0 means deactivated
    alibi_slopes=None,
    deterministic=False,
    return_attn_probs=False,
    block_table=None,
    *,
    out=None,
):
    return FlashAttnVarlenFunc.apply(
        q,
        k,
        v,
        cu_seqlens_q,
        cu_seqlens_k,
        max_seqlen_q,
        max_seqlen_k,
        dropout_p,
        softmax_scale,
        causal,
        window_size,
        softcap,
        alibi_slopes,
        deterministic,
        return_attn_probs,
        block_table,
        out,
    )


def flash_attn_with_kvcache(
    q,
    k_cache,
    v_cache,
    k=None,
    v=None,
    rotary_cos=None,
    rotary_sin=None,
    cache_seqlens: Optional[Union[(int, torch.Tensor)]] = None,
    cache_batch_idx: Optional[torch.Tensor] = None,
    block_table: Optional[torch.Tensor] = None,
    softmax_scale=None,
    causal=False,
    window_size=(-1, -1),  # -1 means infinite context window
    softcap=0.0, # 0.0 means deactivated
    rotary_interleaved=True,
    alibi_slopes=None,
    num_splits=0,
    return_softmax_lse=False,
    *,
    out=None,
):
    assert k_cache.stride(-1) == 1, (
        "k_cache must have contiguous last dimension"
    )
    assert v_cache.stride(-1) == 1, (
        "v_cache must have contiguous last dimension"
    )
    q, k, v = [maybe_contiguous(x) for x in (q, k, v)]
    if softmax_scale is None:
        softmax_scale = q.shape[-1] ** (-0.5)
    if cache_seqlens is not None and isinstance(cache_seqlens, int):
        cache_seqlens = torch.full(
            (k_cache.shape[0],), cache_seqlens,
            dtype=torch.int32, device=k_cache.device
        )
        cache_seqlens = maybe_contiguous(cache_seqlens)
    cache_batch_idx = maybe_contiguous(cache_batch_idx)
    block_table = maybe_contiguous(block_table)
    out, softmax_lse = ops.fwd_kvcache(
        q=q,
        kcache=k_cache,
        vcache=v_cache,
        k=k,
        v=v,
        seqlens_k=cache_seqlens,
        rotary_cos=rotary_cos,
        rotary_sin=rotary_sin,
        cache_batch_idx=cache_batch_idx,
        block_table=block_table,
        alibi_slopes=alibi_slopes,
        out=out,
        softmax_scale=softmax_scale,
        causal=causal,
        window_size_left=window_size[0],
        window_size_right=window_size[1],
        softcap=softcap,
        rotary_interleaved=rotary_interleaved,
        num_splits=num_splits,
    )  # type: ignore
    return (out, softmax_lse) if return_softmax_lse else out



import os
import tempfile
from typing import Optional

from loguru import logger
from prometheus_client import REGISTRY, CollectorRegistry, multiprocess

# Global temporary directory for prometheus multiprocessing
_prometheus_multiproc_dir: Optional[tempfile.TemporaryDirectory] = None


def setup_multiprocess_prometheus():
    """Set up prometheus multiprocessing directory if not already configured.

    """
    global _prometheus_multiproc_dir

    if "PROMETHEUS_MULTIPROC_DIR" not in os.environ:
        # Make TemporaryDirectory for prometheus multiprocessing
        # Note: global TemporaryDirectory will be automatically
        # cleaned up upon exit.
        _prometheus_multiproc_dir = tempfile.TemporaryDirectory()
        os.environ["PROMETHEUS_MULTIPROC_DIR"] = _prometheus_multiproc_dir.name
        logger.debug("Created PROMETHEUS_MULTIPROC_DIR at {}",
                     _prometheus_multiproc_dir.name)
    else:
        logger.warning("Found PROMETHEUS_MULTIPROC_DIR was set by user. "
                       "This directory must be wiped between Aphrodite runs or "
                       "you will find inaccurate metrics. Unset the variable "
                       "and Aphrodite will properly handle cleanup.")


def get_prometheus_registry():
    """Get the appropriate prometheus registry based on multiprocessing 
    configuration.

    Returns:
        Registry: A prometheus registry
    """
    if os.getenv("PROMETHEUS_MULTIPROC_DIR") is not None:
        logger.debug("Using multiprocess registry for prometheus metrics")
        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
        return registry

    return REGISTRY


def unregister_aphrodite_metrics():
    """Unregister any existing Aphrodite collectors from the prometheus registry.

    This is useful for testing and CI/CD where metrics may be registered
    multiple times across test runs.

    Also, in case of multiprocess, we need to unregister the metrics from the 
    global registry.
    """
    registry = REGISTRY
    # Unregister any existing Aphrodite collectors
    for collector in list(registry._collector_to_names):
        if hasattr(collector, "_name") and "aphrodite" in collector._name:
            registry.unregister(collector)


def shutdown_prometheus():
    """Shutdown prometheus metrics."""

    path = _prometheus_multiproc_dir
    if path is None:
        return
    try:
        pid = os.getpid()
        multiprocess.mark_process_dead(pid, path)
        logger.debug("Marked Prometheus metrics for process {} as dead", pid)
    except Exception as e:
        logger.error("Error during metrics cleanup: {}", str(e))



###############################################################################
# Copyright (C) 2024-2025 Habana Labs, Ltd. an Intel Company
#
# This source code is licensed under the Apache 2.0 license found in the
# LICENSE file in the root directory of this source tree.
###############################################################################
import math
from typing import Optional, Tuple

import habana_frameworks.torch as htorch
import habana_frameworks.torch.core as htcore
import torch
import torch.nn.functional as F

from aphrodite.hpu_extension.flags import enabled_flags


def grouped_max(block_max, batch_size, block_groups):
    group_max = torch.full([batch_size + 1, *block_max.shape[1:]],
                           -math.inf,
                           dtype=block_max.dtype,
                           device=block_max.device)
    group_max = group_max.index_reduce_(0, block_groups, block_max, 'amax')
    group_max = group_max.index_select(0, block_groups)
    return group_max


def b2b_impl(tensor, block_mapping, matmul_op):
    shape = tuple(tensor.shape)
    return matmul_op(block_mapping, tensor.view(shape[0],
                                                -1)).view(-1, *shape[1:])


def batch2block(tensor, block_mapping, matmul_op=torch.matmul):
    return b2b_impl(tensor, block_mapping, matmul_op)


def block2batch(tensor, block_mapping, matmul_op=torch.matmul):
    return b2b_impl(tensor, block_mapping.t(), matmul_op)


def pipelined_pa(attn, value, block_groups, block_mapping, batch_size,
                 matmul_av_op, batch2block_matmul_op, block2batch_matmul_op):
    # When fp32_softmax is enabled attn is left in fp32 after Q@K
    # We can return to native dtype after we renormalize and calculate the adjustments

    # Normalize the attention scores and cast attn to native dtype
    block_max = attn.amax(dim=-1, keepdim=True)
    adjustment_target_shape = block_max.shape
    attn = attn.sub(block_max)
    attn = attn.exp()
    attn = attn.to(value.dtype)
    block_sums = attn.sum(dim=-1, keepdim=True)
    attn = matmul_av_op(attn, value)
    block_max = block_max.squeeze()
    block_sums = block_sums.squeeze()

    # Calculate maximum of blocks that belong to the same sequences
    # and cast adjustments to native dtype
    group_max = grouped_max(block_max, batch_size, block_groups)
    block_adjustment = (block_max - group_max).exp()
    block_adjustment = block_adjustment.to(value.dtype)
    sum_adjusted = block_sums.mul(block_adjustment)

    # Sum block's sums that belongs to the same sequences
    group_sum_adjusted = block2batch(sum_adjusted, block_mapping,
                                     block2batch_matmul_op)
    group_sum_adjusted = batch2block(group_sum_adjusted, block_mapping,
                                     batch2block_matmul_op)
    sum_adjusted = sum_adjusted.view(*adjustment_target_shape)
    group_sum_adjusted = group_sum_adjusted.view(*adjustment_target_shape)
    block_adjustment = block_adjustment.view(*adjustment_target_shape)

    # For stability in case some of the sums have been zeroed out during block aggretation
    group_sum_adjusted = torch.maximum(group_sum_adjusted, sum_adjusted)

    # Post processing for the attention scores
    rescale = block_adjustment.div(group_sum_adjusted)
    attn = attn.mul(rescale)
    return attn


def flat_pa(query, key_cache, value_cache, block_list, block_mapping,
            block_bias, block_groups, scale, matmul_qk_op, matmul_av_op,
            batch2block_matmul_op, block2batch_matmul_op, keys_fetch_func,
            values_fetch_func, **ignored_args):
    batch_size, _, hidden_size = query.shape
    _, _, kv_heads, head_size = key_cache.shape
    q_heads = hidden_size // head_size

    query_shape = (-1, q_heads, 1, head_size)
    query = batch2block(scale * query, block_mapping,
                        batch2block_matmul_op).view(query_shape)
    key = keys_fetch_func(key_cache, block_list).transpose(1, 2)
    value = values_fetch_func(value_cache, block_list).transpose(1, 2)
    block_bias = block_bias.view(key.size(0), 1, 1, -1)
    if kv_heads != q_heads:
        block_bias = block_bias.unsqueeze(1)
        query = query.unflatten(1, (kv_heads, -1))
        key = key.unflatten(1, (kv_heads, 1))
        value = value.unflatten(1, (kv_heads, 1))
        key = key.transpose(3, 4)
    else:
        key = key.transpose(2, 3)

    attn = matmul_qk_op(query, key)
    if 'fp32_softmax' in enabled_flags():
        attn = attn.float()
        htcore.mark_step()
    attn = attn + block_bias
    attn = pipelined_pa(attn,
                        value,
                        block_groups,
                        block_mapping,
                        batch_size=batch_size,
                        matmul_av_op=matmul_av_op,
                        batch2block_matmul_op=batch2block_matmul_op,
                        block2batch_matmul_op=block2batch_matmul_op)
    attn = block2batch(attn, block_mapping, block2batch_matmul_op)
    attn = attn.squeeze(-2)
    if kv_heads != q_heads:
        attn = attn.flatten(1, 2)
    return attn


def _flex_prompt_attention(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    scale: float,
    **ignored_args,
) -> torch.Tensor:
    query = query.transpose(1, 2)
    key = key.transpose(1, 2)
    value = value.transpose(1, 2)

    def _causal(
        score: torch.Tensor,
        batch: torch.Tensor,
        head: torch.Tensor,
        token_q: torch.Tensor,
        token_kv: torch.Tensor,
    ) -> torch.Tensor:
        return torch.where(token_q >= token_kv, score, float("-inf"))

    from torch.nn.attention.flex_attention import flex_attention

    attn_weights = flex_attention(
        query,
        key,
        value,
        score_mod=_causal,
        enable_gqa=True,
        return_lse=False,
        block_mask=None,
        scale=scale,
    )

    attn_weights = attn_weights.transpose(1, 2)
    return attn_weights


def _naive_prompt_attention(query: torch.Tensor,
                            key: torch.Tensor,
                            value: torch.Tensor,
                            scale: float,
                            attn_bias: Optional[torch.Tensor] = None,
                            matmul_qk_op=torch.matmul,
                            softmax_op=torch.softmax,
                            matmul_av_op=torch.matmul,
                            **ignored_args) -> torch.Tensor:
    query = query.transpose(1, 2)
    key = key.transpose(1, 2)
    value = value.transpose(1, 2)
    query_heads = query.size(1)
    kv_heads = key.size(1)
    if query_heads != kv_heads:
        query = query.unflatten(1, (kv_heads, -1))
        key = key.unflatten(1, (kv_heads, 1))
        value = value.unflatten(1, (kv_heads, 1))
        if attn_bias is not None:
            attn_bias = attn_bias.unsqueeze(2)
    attn_weights = matmul_qk_op(query * scale, key.transpose(-1, -2))
    if 'fp32_softmax' in enabled_flags():
        softmax_op = torch.softmax
        attn_weights = attn_weights.float()
        htcore.mark_step()
    if attn_bias is not None:
        attn_weights = attn_weights.add(attn_bias)
    attn_weights = softmax_op(attn_weights, dim=-1)
    attn_weights = attn_weights.to(query.dtype)
    attn_weights = matmul_av_op(attn_weights, value)
    if query_heads != kv_heads:
        attn_weights = attn_weights.flatten(1, 2)
    attn_weights = attn_weights.transpose(1, 2)
    return attn_weights


def _fsdpa_prompt_attention(query: torch.Tensor,
                            key: torch.Tensor,
                            value: torch.Tensor,
                            scale: float,
                            fsdpa_op,
                            is_causal: bool,
                            attn_bias: Optional[torch.Tensor] = None,
                            valid_seq_lengths: Optional[torch.Tensor] = None,
                            **ignored_args) -> torch.Tensor:
    query = query.transpose(1, 2)
    key = key.transpose(1, 2)
    value = value.transpose(1, 2)
    if 'fp32_softmax' in enabled_flags():
        softmax_mode = 'fp32'
    else:
        softmax_mode = 'fast'
    recompute_mode = True
    assert attn_bias is not None or valid_seq_lengths is not None, \
        'Either attn_bias or valid_seq_lengths must be != None'
    if is_causal and attn_bias is not None:
        # TODO: causal + attn_bias is not yet supported
        is_causal = False
        valid_seq_lengths = None
    attn_weights = fsdpa_op(query, key, value, attn_bias, 0.0, is_causal,
                            scale, softmax_mode, recompute_mode,
                            valid_seq_lengths, 'right')
    attn_weights = attn_weights.transpose(1, 2)
    return attn_weights


def prompt_attention(
    impl: str,
    **args,
) -> torch.Tensor:
    _get_context(args)
    impl_mapping = {
        'naive': _naive_prompt_attention,
        'fsdpa': _fsdpa_prompt_attention,
        'flex': _flex_prompt_attention,
    }
    assert impl in impl_mapping, f'Unsupported implementation: {impl}'
    return impl_mapping[impl](**args)


def _get_all(data, *keys):
    return [data.get(k, None) for k in keys]


def _include_past(tensor_str, fn_str, cache_str, args):
    all_tensors = _get_all(args, tensor_str, fn_str, cache_str, 'block_list')
    if all(t is not None for t in all_tensors):
        current, fn, cache, block_list = all_tensors
        past = fn(cache, block_list)
        past = past.reshape(current.size(0), -1, past.shape[2], past.shape[3])
        current = torch.concat((past, current), dim=1)
        args[tensor_str] = current


def _get_context(args):
    _include_past('key', 'keys_fetch_func', 'key_cache', args)
    _include_past('value', 'values_fetch_func', 'value_cache', args)


class LoraMask:
    lora_mask = None

    @staticmethod
    def setLoraMask(mask):
        LoraMask.lora_mask = mask

    @staticmethod
    def getLoraMask():
        return LoraMask.lora_mask


def dispatch_bgmv_linear(
    y: torch.Tensor,
    x: torch.Tensor,
    wa_t_all: torch.Tensor,
    wb_t_all: torch.Tensor,
    layer_idx: int,
    scale: float,
):
    """
    `wa_t_all` and `wb_t_all` contains all LoRA A and LoRA B weight matrices
    stacked at dimension 0 into single tensors, assuming same rank. `wa` is the
    reshaped and transposed version of `wa_t_all` of shape
    (h_in, max_loras * lora_rank) and `wb` is the transposed and reshaped
    version of `wb_t_all` of shape (max_loras * lora_rank, h_out).

    Matmul input `x` with `wa`. Multiply `x` with a mask to zero-out inputs of
    inactive LoRA indices. Matmul masked output with `wb` and scale it to get
    the final output.
    """

    assert layer_idx == 0, f'layer_idx should be 0, but got {layer_idx}'
    mask = LoraMask.getLoraMask()

    wa = wa_t_all[:, 0, :, :]
    wb = wb_t_all[:, 0, :, :].transpose(1, 2)
    wa = wa.reshape(wa.shape[0] * wa.shape[1], wa.shape[2]).transpose(0, 1)
    wb = wb.reshape(wb.shape[0] * wb.shape[1], wb.shape[2])

    out = x @ wa
    assert (out.shape == mask.shape)
    out = out * mask
    out = out @ wb
    y += out * scale


def dispatch_bgmv_embedding(
    y: torch.Tensor,
    x: torch.Tensor,
    wb_t_all: torch.Tensor,
    layer_idx: int,
):
    """
    `wb_t_all` contains all LoRA-B weight matrices stacked at dimension 0 into
    a single tensor, assuming same rank. `wb` is the transposed and reshaped
    version of `wb_t_all` of shape (num_loras * lora_rank, embedding_dim).

    Output of LoRA-A embedding (tensor x) is repeated max_loras times to match
    the shape of `wb`. Multiply `x` with a mask to zero-out inputs of inactive
    LoRA indices. Matmul masked output with `wb` and scale it to get the final
    output.
    """

    assert layer_idx == 0, f'layer_idx should be 0, but got {layer_idx}'
    max_loras = wb_t_all.size(0)

    x = x.repeat(1, max_loras)
    x = x * LoraMask.getLoraMask()
    wb = wb_t_all[:, 0, :, :].transpose(1, 2)
    wb = wb.reshape(wb.shape[0] * wb.shape[1], wb.shape[2])
    out = x @ wb
    y += out


class MoeMatmul(torch.nn.Module):

    def __init__(self):
        super().__init__()

    def set_weight(self, w):
        self.weight = w

    def forward(self, state, expert_id, w):
        raise NotImplementedError()


class AphroditeMixtureOfExpertsOp(torch.nn.Module):

    def __init__(self, num_total_experts):
        super().__init__()
        self.w13_list = torch.nn.ModuleList(
            [MoeMatmul() for _ in range(num_total_experts)])
        self.w2_list = torch.nn.ModuleList(
            [MoeMatmul() for _ in range(num_total_experts)])
        self.num_experts = num_total_experts

    def forward(self,
                hidden_states,
                expert_routing_table,
                router_weights,
                permuted_weights=True,
                activation="silu"):
        # pre-processing for custom op inputs
        experts_range = range(self.num_experts)
        w1_list = [self.w13_list[i].weight.squeeze() for i in experts_range]
        w2_list = [self.w2_list[i].weight.squeeze() for i in experts_range]
        return torch.ops.hpu.mixture_of_experts(
            hidden_states=hidden_states,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w12=w1_list,
            w3=w2_list,
            permuted_weights=permuted_weights,
            activation=activation,
            experts_min=0,
            experts_max=self.num_experts - 1)


class DynamicFusedMOE(torch.nn.Module):

    def __init__(self, num_total_experts):
        super().__init__()
        self.MoeOp = AphroditeMixtureOfExpertsOp(num_total_experts)

    def forward(self, hidden_states, score, topk):
        htorch.core.mark_step()
        routing_weights = F.softmax(score, dim=1, dtype=torch.float32)
        routing_weights, selected_experts = torch.topk(routing_weights,
                                                       topk,
                                                       dim=-1)
        routing_weights /= routing_weights.sum(dim=-1, keepdim=True)
        routing_weights = routing_weights.to(hidden_states.dtype)

        final_hidden_states = self.MoeOp(
            hidden_states=hidden_states,
            expert_routing_table=selected_experts,
            router_weights=routing_weights,
            permuted_weights=True,
            activation="silu",
        )

        return final_hidden_states.view(-1, hidden_states.shape[1])


# fp8
def scaled_fp8_quant(
    input: torch.Tensor,
    scale: Optional[torch.Tensor] = None,
    batch_dim_padding: Optional[int] = None,
    scale_ub: Optional[torch.Tensor] = None,
    use_per_token_if_dynamic: bool = False,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Quantize input tensor to FP8 and return quantized tensor and scale.
    This function supports both static and dynamic quantization: If you
    provide the scale, it will use static scaling and if you omit it,
    the scale will be determined dynamically. The function also allows
    optional padding of the output tensor for downstream kernels that
    will benefit from padding.
    Args:
        input: The input tensor to be quantized to FP8
        scale: Optional scaling factor for the FP8 quantization
        scale_ub: Optional upper bound for scaling factor in dynamic
            per token case
        batch_dim_padding: If specified, pad the first dimension
            of the output to at least this value.
        use_per_token_if_dynamic: Whether to do per_tensor or per_token
            in the dynamic quantization case.
    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The output tensor in FP8 and
            scaling factor.
    """
    if batch_dim_padding:
        shape = (max(batch_dim_padding, input.shape[0]), *input.shape[1:])
        output = torch.empty(shape,
                             device=input.device,
                             dtype=torch.float8_e4m3fn)
    else:
        output = torch.empty_like(input, dtype=torch.float8_e4m3fn)
    if scale is None:
        raise NotImplementedError(
            "dynamic scaled_fp8_quant not implemented for HPU")
        # TODO: calculate scale to match gaudi2 240 range instead of 448
        if use_per_token_if_dynamic:
            scale = torch.empty((input.numel() // input.shape[-1], 1),
                                device=input.device,
                                dtype=torch.float32)
            torch.ops._C.dynamic_per_token_scaled_fp8_quant(
                output, input, scale, scale_ub)
        else:
            scale = torch.zeros(1, device=input.device, dtype=torch.float32)
            torch.ops._C.dynamic_scaled_fp8_quant(output, input, scale)
    else:
        output = torch.ops.hpu.cast_to_fp8_v2(input,
                                              1 / scale,
                                              False,
                                              False,
                                              dtype=torch.float8_e4m3fn)[0]

    return output, scale



"""UNet2DConditionModel implementation for Stable Diffusion."""

import math
from typing import List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from aphrodite.config import AphroditeConfig
from aphrodite.modeling.models.interfaces import SupportsQuant
from aphrodite.modeling.models.utils import maybe_prefix


def get_timestep_embedding(
    timesteps: torch.Tensor,
    embedding_dim: int,
    flip_sin_to_cos: bool = False,
    downscale_freq_shift: float = 1,
    scale: float = 1,
    max_period: int = 10000,
):
    """
    This matches the implementation in Denoising Diffusion Probabilistic
    Models.

    Create sinusoidal timestep embeddings.
    """
    half_dim = embedding_dim // 2
    exponent = -math.log(max_period) * torch.arange(
        start=0, end=half_dim, dtype=torch.float32, device=timesteps.device
    )
    exponent = exponent / (half_dim - downscale_freq_shift)

    emb = torch.exp(exponent)
    emb = timesteps[:, None].float() * emb[None, :]

    # scale embeddings
    emb = scale * emb

    # concat sine and cosine embeddings
    emb = torch.cat([torch.sin(emb), torch.cos(emb)], dim=-1)

    # flip sine and cosine embeddings
    if flip_sin_to_cos:
        emb = torch.cat([emb[:, half_dim:], emb[:, :half_dim]], dim=-1)

    # zero pad
    if embedding_dim % 2 == 1:
        emb = torch.nn.functional.pad(emb, (0, 1, 0, 0))
    return emb


class TimestepEmbedding(nn.Module):
    def __init__(
        self,
        in_channels: int,
        time_embed_dim: int,
        act_fn: str = "silu",
    ):
        super().__init__()

        self.linear_1 = nn.Linear(in_channels, time_embed_dim)
        self.linear_2 = nn.Linear(time_embed_dim, time_embed_dim)

        if act_fn == "silu":
            self.act = F.silu
        else:
            raise ValueError(f"Unsupported activation function: {act_fn}")

    def forward(self, sample):
        sample = self.linear_1(sample)
        sample = self.act(sample)
        sample = self.linear_2(sample)
        return sample


class Downsample2D(nn.Module):
    def __init__(self, channels: int, use_conv: bool = True, padding: int = 1):
        super().__init__()
        self.channels = channels
        self.use_conv = use_conv
        self.padding = padding

        if use_conv:
            self.conv = nn.Conv2d(channels, channels, kernel_size=3,
                                  stride=2, padding=padding)
        else:
            self.conv = None

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        if self.use_conv:
            hidden_states = self.conv(hidden_states)
        else:
            hidden_states = F.avg_pool2d(hidden_states, kernel_size=2, stride=2)
        return hidden_states


class Upsample2D(nn.Module):
    def __init__(self, channels: int, use_conv: bool = True):
        super().__init__()
        self.channels = channels
        self.use_conv = use_conv

        if use_conv:
            self.conv = nn.Conv2d(channels, channels, kernel_size=3, padding=1)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        # Upsample by factor of 2
        hidden_states = F.interpolate(hidden_states, scale_factor=2.0, mode="nearest")

        if self.use_conv:
            hidden_states = self.conv(hidden_states)

        return hidden_states


class UNetResidualBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        time_embedding_dim: int = 1280,
        groups: int = 32,
        eps: float = 1e-6,
        use_shortcut: bool = True,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.use_shortcut = use_shortcut

        # First conv block
        self.norm1 = nn.GroupNorm(groups, in_channels, eps=eps)
        self.conv1 = nn.Conv2d(
            in_channels, out_channels, kernel_size=3, padding=1)

        # Time embedding projection
        self.time_emb_proj = nn.Linear(time_embedding_dim, out_channels)

        # Second conv block
        self.norm2 = nn.GroupNorm(groups, out_channels, eps=eps)
        self.conv2 = nn.Conv2d(
            out_channels, out_channels, kernel_size=3, padding=1)

        # Shortcut connection
        if self.use_shortcut and in_channels != out_channels:
            self.conv_shortcut = nn.Conv2d(
                in_channels, out_channels, kernel_size=1)
        else:
            self.conv_shortcut = None

    def forward(
        self,
        hidden_states: torch.Tensor,
        time_emb: torch.Tensor,
    ) -> torch.Tensor:
        residual = hidden_states

        # First conv block
        hidden_states = self.norm1(hidden_states)
        hidden_states = F.silu(hidden_states)
        hidden_states = self.conv1(hidden_states)

        # Time embedding
        time_emb = F.silu(time_emb)
        time_emb = self.time_emb_proj(time_emb)
        # Reshape time embedding to match spatial dimensions
        time_emb = time_emb[:, :, None, None]
        hidden_states = hidden_states + time_emb

        # Second conv block
        hidden_states = self.norm2(hidden_states)
        hidden_states = F.silu(hidden_states)
        hidden_states = self.conv2(hidden_states)

        # Shortcut connection
        if self.use_shortcut:
            if self.conv_shortcut is not None:
                residual = self.conv_shortcut(residual)
            hidden_states = hidden_states + residual

        return hidden_states


class UNetAttention(nn.Module):
    """Attention layer matching Diffusers structure exactly."""

    def __init__(
        self,
        channels: int,
        cross_attention_dim: Optional[int] = None,
        num_attention_heads: int = 8,
        attention_head_dim: int = None,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()
        self.channels = channels
        self.cross_attention_dim = cross_attention_dim or channels

        if attention_head_dim is None:
            attention_head_dim = channels // num_attention_heads

        self.num_attention_heads = num_attention_heads
        self.attention_head_dim = attention_head_dim
        self.inner_dim = num_attention_heads * attention_head_dim
        self.scale = 1.0 / math.sqrt(attention_head_dim)

        # Query projection (always from input features)
        self.to_q = nn.Linear(channels, self.inner_dim, bias=False)

        # Key and Value projections (from cross_attention_dim for cross-attn)
        self.to_k = nn.Linear(
            self.cross_attention_dim, self.inner_dim, bias=False)
        self.to_v = nn.Linear(
            self.cross_attention_dim, self.inner_dim, bias=False)

        # Output projection
        self.to_out = nn.ModuleList([
            nn.Linear(self.inner_dim, channels),
            nn.Identity(),  # Dropout placeholder
        ])

    def forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        batch_size, seq_len, channels = hidden_states.shape

        # Use encoder_hidden_states for cross-attention, hidden_states for
        # self-attention
        context = (
            encoder_hidden_states
            if encoder_hidden_states is not None
            else hidden_states
        )

        # Get QKV
        q = self.to_q(hidden_states)
        k = self.to_k(context)
        v = self.to_v(context)

        # Reshape for multi-head attention
        q = q.view(
            batch_size, seq_len, self.num_attention_heads,
            self.attention_head_dim,
        )
        k = k.view(
            batch_size, context.shape[1], self.num_attention_heads,
            self.attention_head_dim,
        )
        v = v.view(
            batch_size, context.shape[1], self.num_attention_heads,
            self.attention_head_dim,
        )

        # Transpose for attention: (batch, heads, seq_len, head_dim)
        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)

        # Compute attention
        attn_weights = torch.matmul(q, k.transpose(-2, -1)) * self.scale
        attn_weights = F.softmax(attn_weights, dim=-1)
        attn_output = torch.matmul(attn_weights, v)

        # Reshape back
        attn_output = attn_output.transpose(1, 2).contiguous()
        attn_output = attn_output.view(batch_size, seq_len, self.inner_dim)

        # Output projection
        attn_output = self.to_out[0](attn_output)
        attn_output = self.to_out[1](attn_output)

        return attn_output


class FeedForward(nn.Module):
    """GeGLU feedforward network matching Diffusers structure."""

    def __init__(
        self,
        dim: int,
        dim_out: Optional[int] = None,
        mult: int = 4,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()
        inner_dim = int(dim * mult)
        dim_out = dim_out or dim

        # GeGLU: project to 2x inner_dim, then split for gate and value
        self.net = nn.ModuleList([
            nn.ModuleList([
                nn.Linear(dim, inner_dim * 2),  # proj layer (GeGLU splits this)  # noqa: E501
            ]),
            nn.Identity(),  # Placeholder for potential dropout
            nn.Linear(inner_dim, dim_out),
        ])

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        # GeGLU activation
        hidden_states, gate = self.net[0][0](hidden_states).chunk(2, dim=-1)
        hidden_states = hidden_states * F.gelu(gate)

        # Dropout (currently identity)
        hidden_states = self.net[1](hidden_states)

        # Final projection
        hidden_states = self.net[2](hidden_states)

        return hidden_states


class BasicTransformerBlock(nn.Module):
    """Basic transformer block matching Diffusers structure exactly."""

    def __init__(
        self,
        dim: int,
        cross_attention_dim: int = 768,
        num_attention_heads: int = 8,
        attention_head_dim: int = None,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()

        # Layer norms
        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.norm3 = nn.LayerNorm(dim)

        # Self-attention
        self.attn1 = UNetAttention(
            channels=dim,
            cross_attention_dim=None,  # Self-attention
            num_attention_heads=num_attention_heads,
            attention_head_dim=attention_head_dim,
            quant_config=quant_config,
            prefix=maybe_prefix(prefix, "attn1"),
        )

        # Cross-attention
        self.attn2 = UNetAttention(
            channels=dim,
            cross_attention_dim=cross_attention_dim,
            num_attention_heads=num_attention_heads,
            attention_head_dim=attention_head_dim,
            quant_config=quant_config,
            prefix=maybe_prefix(prefix, "attn2"),
        )

        # Feedforward
        self.ff = FeedForward(
            dim=dim,
            # 4x expansion for GeGLU (4x * 2 = 8x total, matches 2560 output
            # for 320 input)
            mult=4,
            quant_config=quant_config,
            prefix=maybe_prefix(prefix, "ff"),
        )

    def forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
    ) -> torch.Tensor:
        # Self-attention
        residual = hidden_states
        hidden_states = self.norm1(hidden_states)
        hidden_states = self.attn1(hidden_states)
        hidden_states = hidden_states + residual

        # Cross-attention
        residual = hidden_states
        hidden_states = self.norm2(hidden_states)
        hidden_states = self.attn2(hidden_states, encoder_hidden_states)
        hidden_states = hidden_states + residual

        # Feedforward
        residual = hidden_states
        hidden_states = self.norm3(hidden_states)
        hidden_states = self.ff(hidden_states)
        hidden_states = hidden_states + residual

        return hidden_states


class Transformer2DModel(nn.Module):
    """2D Transformer model matching Diffusers structure."""

    def __init__(
        self,
        num_attention_heads: int,
        attention_head_dim: int,
        in_channels: int,
        cross_attention_dim: int = 768,
        num_layers: int = 1,
        norm_num_groups: int = 32,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()
        self.num_attention_heads = num_attention_heads
        self.attention_head_dim = attention_head_dim
        self.in_channels = in_channels
        self.inner_dim = num_attention_heads * attention_head_dim

        # Input/output projections
        self.norm = nn.GroupNorm(norm_num_groups, in_channels, eps=1e-6)
        self.proj_in = nn.Conv2d(in_channels, self.inner_dim, kernel_size=1)
        self.proj_out = nn.Conv2d(self.inner_dim, in_channels, kernel_size=1)

        # Transformer blocks
        self.transformer_blocks = nn.ModuleList([
            BasicTransformerBlock(
                dim=self.inner_dim,
                cross_attention_dim=cross_attention_dim,
                num_attention_heads=num_attention_heads,
                attention_head_dim=attention_head_dim,
                quant_config=quant_config,
                prefix=maybe_prefix(prefix, f"transformer_blocks.{i}"),
            )
            for i in range(num_layers)
        ])

    def forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
    ) -> torch.Tensor:
        batch_size, channels, height, width = hidden_states.shape
        residual = hidden_states

        # Input projection
        hidden_states = self.norm(hidden_states)
        hidden_states = self.proj_in(hidden_states)

        # Reshape to sequence format
        inner_dim = hidden_states.shape[1]
        hidden_states = hidden_states.view(
            batch_size, inner_dim, height * width,
        )
        hidden_states = hidden_states.transpose(1, 2)

        # Apply transformer blocks
        for block in self.transformer_blocks:
            hidden_states = block(hidden_states, encoder_hidden_states)

        # Reshape back to spatial format
        hidden_states = hidden_states.transpose(1, 2)
        hidden_states = hidden_states.view(
            batch_size, inner_dim, height, width)

        # Output projection
        hidden_states = self.proj_out(hidden_states)

        return hidden_states + residual


class UNetAttentionBlock(nn.Module):
    """Attention block wrapper using Transformer2DModel to match Diffusers
    exactly."""

    def __init__(
        self,
        channels: int,
        cross_attention_dim: int = 768,
        num_attention_heads: int = 8,
        attention_head_dim: int = None,
        groups: int = 32,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()

        if attention_head_dim is None:
            attention_head_dim = channels // num_attention_heads

        # Use Transformer2DModel to match Diffusers structure exactly
        self.transformer = Transformer2DModel(
            num_attention_heads=num_attention_heads,
            attention_head_dim=attention_head_dim,
            in_channels=channels,
            cross_attention_dim=cross_attention_dim,
            num_layers=1,
            norm_num_groups=groups,
            quant_config=quant_config,
            prefix=prefix,
        )

    def forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
    ) -> torch.Tensor:
        return self.transformer(hidden_states, encoder_hidden_states)


class CrossAttnDownBlock2D(nn.Module):
    """Cross-attention down block matching Diffusers structure."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        time_embedding_dim: int,
        cross_attention_dim: int = 768,
        num_attention_heads: int = 8,
        attention_head_dim: int = None,
        num_layers: int = 2,
        add_downsample: bool = True,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()

        if attention_head_dim is None:
            attention_head_dim = out_channels // num_attention_heads

        # Residual blocks
        self.resnets = nn.ModuleList()
        for i in range(num_layers):
            in_ch = in_channels if i == 0 else out_channels
            self.resnets.append(
                UNetResidualBlock(
                    in_channels=in_ch,
                    out_channels=out_channels,
                    time_embedding_dim=time_embedding_dim,
                )
            )

        # Attention blocks (one per resnet)
        self.attentions = nn.ModuleList()
        for i in range(num_layers):
            self.attentions.append(
                Transformer2DModel(
                    num_attention_heads=num_attention_heads,
                    attention_head_dim=attention_head_dim,
                    in_channels=out_channels,
                    cross_attention_dim=cross_attention_dim,
                    num_layers=1,
                    prefix=maybe_prefix(prefix, f"attentions.{i}"),
                )
            )

        # Downsampler
        if add_downsample:
            self.downsamplers = nn.ModuleList([Downsample2D(out_channels)])
        else:
            self.downsamplers = None

    def forward(
        self,
        hidden_states: torch.Tensor,
        time_emb: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        output_states = []

        for resnet, attn in zip(self.resnets, self.attentions):
            hidden_states = resnet(hidden_states, time_emb)
            hidden_states = attn(hidden_states, encoder_hidden_states)
            output_states.append(hidden_states)

        if self.downsamplers is not None:
            hidden_states = self.downsamplers[0](hidden_states)
            output_states.append(hidden_states)

        return hidden_states, output_states


class DownBlock2D(nn.Module):
    """Down block without attention matching Diffusers structure."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        time_embedding_dim: int,
        num_layers: int = 2,
        add_downsample: bool = True,
        prefix: str = "",
    ):
        super().__init__()

        # Residual blocks
        self.resnets = nn.ModuleList()
        for i in range(num_layers):
            in_ch = in_channels if i == 0 else out_channels
            self.resnets.append(
                UNetResidualBlock(
                    in_channels=in_ch,
                    out_channels=out_channels,
                    time_embedding_dim=time_embedding_dim,
                )
            )

        # Downsampler
        if add_downsample:
            self.downsamplers = nn.ModuleList([Downsample2D(out_channels)])
        else:
            self.downsamplers = None

    def forward(
        self,
        hidden_states: torch.Tensor,
        time_emb: torch.Tensor,
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        output_states = []

        for resnet in self.resnets:
            hidden_states = resnet(hidden_states, time_emb)
            output_states.append(hidden_states)

        if self.downsamplers is not None:
            hidden_states = self.downsamplers[0](hidden_states)
            output_states.append(hidden_states)

        return hidden_states, output_states


class CrossAttnUpBlock2D(nn.Module):
    """Cross-attention up block matching Diffusers structure."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        prev_output_channel: int,
        time_embedding_dim: int,
        cross_attention_dim: int = 768,
        num_attention_heads: int = 8,
        attention_head_dim: int = None,
        num_layers: int = 3,
        add_upsample: bool = True,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()

        if attention_head_dim is None:
            attention_head_dim = out_channels // num_attention_heads

        # Residual blocks with exact Diffusers channel calculations
        self.resnets = nn.ModuleList()

        if out_channels == 1280:  # up_blocks[1]
            resnet_input_channels = [2560, 2560, 1920]
        elif out_channels == 640:  # up_blocks[2]
            resnet_input_channels = [1920, 1280, 960]
        elif out_channels == 320:  # up_blocks[3]
            resnet_input_channels = [960, 640, 640]
        else:
            # Fallback calculation
            skip_channels = self._get_skip_channels(out_channels, num_layers)
            resnet_input_channels = []
            for i in range(num_layers):
                if i == 0:
                    resnet_in_channels = prev_output_channel + skip_channels[i]
                else:
                    resnet_in_channels = out_channels + skip_channels[i]
                resnet_input_channels.append(resnet_in_channels)

        for i, resnet_in_channels in enumerate(resnet_input_channels):
            self.resnets.append(
                UNetResidualBlock(
                    in_channels=resnet_in_channels,
                    out_channels=out_channels,
                    time_embedding_dim=time_embedding_dim,
                )
            )

        # Attention blocks (one per resnet)
        self.attentions = nn.ModuleList()
        for i in range(num_layers):
            self.attentions.append(
                Transformer2DModel(
                    num_attention_heads=num_attention_heads,
                    attention_head_dim=attention_head_dim,
                    in_channels=out_channels,
                    cross_attention_dim=cross_attention_dim,
                    num_layers=1,
                    prefix=maybe_prefix(prefix, f"attentions.{i}"),
                )
            )

        # Upsampler
        if add_upsample:
            self.upsamplers = nn.ModuleList([Upsample2D(out_channels)])
        else:
            self.upsamplers = None

    def _get_skip_channels(
        self,
        out_channels: int,
        num_layers: int,
    ) -> list[int]:
        """Get the expected skip channel counts to match Diffusers exactly."""
        # These patterns are from exact Diffusers analysis
        if out_channels == 1280:  # up_blocks[1]
            return [1280, 1280, 640]
        elif out_channels == 640:  # up_blocks[2]
            return [1280, 640, 320]
        elif out_channels == 320:  # up_blocks[3]
            return [640, 320, 320]
        else:
            # Fallback for other configurations
            return [out_channels] * num_layers

    def forward(
        self,
        hidden_states: torch.Tensor,
        res_hidden_states_tuple: Tuple[torch.Tensor, ...],
        time_emb: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
    ) -> torch.Tensor:
        # Convert tuple to list for popping (matching Diffusers implementation)
        res_hidden_states_list = list(res_hidden_states_tuple)

        for resnet, attn in zip(self.resnets, self.attentions):
            # Pop skip connection from the end
            # (LIFO - matches Diffusers exactly)
            res_hidden_states = res_hidden_states_list.pop()

            # Concatenate input with skip connection
            hidden_states = torch.cat(
                [hidden_states, res_hidden_states], dim=1)

            # Apply ResNet and attention
            hidden_states = resnet(hidden_states, time_emb)
            hidden_states = attn(hidden_states, encoder_hidden_states)

        # Upsample at the end (matches Diffusers implementation)
        if self.upsamplers is not None:
            hidden_states = self.upsamplers[0](hidden_states)

        return hidden_states


class UpBlock2D(nn.Module):
    """Up block without attention matching Diffusers structure."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        prev_output_channel: int,
        time_embedding_dim: int,
        num_layers: int = 3,
        add_upsample: bool = True,
        prefix: str = "",
    ):
        super().__init__()

        # Residual blocks with exact Diffusers channel calculations for
        # UpBlock2D (up_blocks[0])
        self.resnets = nn.ModuleList()

        # up_blocks[0] has all resnets with 2560 input channels
        # (from Diffusers analysis)
        resnet_input_channels = [2560, 2560, 2560]

        for i, resnet_in_channels in enumerate(resnet_input_channels):
            self.resnets.append(
                UNetResidualBlock(
                    in_channels=resnet_in_channels,
                    out_channels=out_channels,
                    time_embedding_dim=time_embedding_dim,
                )
            )

        # Upsampler
        if add_upsample:
            self.upsamplers = nn.ModuleList([Upsample2D(out_channels)])
        else:
            self.upsamplers = None

    def forward(
        self,
        hidden_states: torch.Tensor,
        res_hidden_states_tuple: Tuple[torch.Tensor, ...],
        time_emb: torch.Tensor,
    ) -> torch.Tensor:
        # Convert tuple to list for popping (matching Diffusers implementation)
        res_hidden_states_list = list(res_hidden_states_tuple)

        for resnet in self.resnets:
            # Pop skip connection from the end
            # (LIFO - matches Diffusers exactly)
            res_hidden_states = res_hidden_states_list.pop()

            # Concatenate input with skip connection
            hidden_states = torch.cat(
                [hidden_states, res_hidden_states], dim=1)

            # Apply ResNet
            hidden_states = resnet(hidden_states, time_emb)

        # Upsample at the end (matches Diffusers implementation)
        if self.upsamplers is not None:
            hidden_states = self.upsamplers[0](hidden_states)

        return hidden_states


class UNetMidBlock2DCrossAttn(nn.Module):
    """Mid block with cross-attention matching Diffusers structure."""

    def __init__(
        self,
        in_channels: int,
        time_embedding_dim: int,
        cross_attention_dim: int = 768,
        num_attention_heads: int = 8,
        attention_head_dim: int = None,
        quant_config=None,
        prefix: str = "",
    ):
        super().__init__()

        if attention_head_dim is None:
            attention_head_dim = in_channels // num_attention_heads

        # 2 resnets with 1 attention in between
        self.resnets = nn.ModuleList([
            UNetResidualBlock(
                in_channels=in_channels,
                out_channels=in_channels,
                time_embedding_dim=time_embedding_dim,
            ),
            UNetResidualBlock(
                in_channels=in_channels,
                out_channels=in_channels,
                time_embedding_dim=time_embedding_dim,
            ),
        ])

        self.attentions = nn.ModuleList([
            Transformer2DModel(
                num_attention_heads=num_attention_heads,
                attention_head_dim=attention_head_dim,
                in_channels=in_channels,
                cross_attention_dim=cross_attention_dim,
                num_layers=1,
                prefix=maybe_prefix(prefix, "attentions.0"),
            )
        ])

    def forward(
        self,
        hidden_states: torch.Tensor,
        time_emb: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
    ) -> torch.Tensor:
        hidden_states = self.resnets[0](hidden_states, time_emb)
        hidden_states = self.attentions[0](
            hidden_states, encoder_hidden_states)
        hidden_states = self.resnets[1](hidden_states, time_emb)

        return hidden_states


class UNet2DConditionModel(nn.Module, SupportsQuant):
    """
    UNet model for conditional image generation (Stable Diffusion).

    This is a 2D UNet model that takes a noisy sample, conditional state, and
    timestep to produce a denoised output.
    """

    is_unet_model = True

    def __init__(
        self,
        *,
        aphrodite_config: AphroditeConfig,
        prefix: str = "",
    ) -> None:
        super().__init__()

        config = aphrodite_config.model_config.hf_config
        quant_config = aphrodite_config.quant_config

        # Configuration
        self.in_channels = config.in_channels
        self.out_channels = config.out_channels
        self.block_out_channels = config.block_out_channels
        self.layers_per_block = config.layers_per_block
        self.attention_head_dim = config.attention_head_dim
        self.cross_attention_dim = config.cross_attention_dim
        self.norm_num_groups = config.norm_num_groups
        # Time embedding dimension (Diffusers uses block_out_channels[0] * 4)
        self.time_embedding_dim = (
            config.time_embedding_dim
            or (self.block_out_channels[0] * 4)
        )

        # Time embedding
        time_embed_dim = self.time_embedding_dim
        self.time_proj = lambda x: get_timestep_embedding(
            x, 320, flip_sin_to_cos=True,
        )
        self.time_embedding = TimestepEmbedding(320, time_embed_dim)

        # Input convolution
        self.conv_in = nn.Conv2d(
            self.in_channels, self.block_out_channels[0],
            kernel_size=3, padding=1,
        )

        # Build down blocks matching Diffusers structure exactly
        self.down_blocks = nn.ModuleList([])

        # down_block_types: ['CrossAttnDownBlock2D', 'CrossAttnDownBlock2D',
        # 'CrossAttnDownBlock2D', 'DownBlock2D']
        down_block_types = config.down_block_types

        for i, (down_block_type, out_channels) in enumerate(
            zip(down_block_types, self.block_out_channels)
        ):
            in_channels = (
                self.block_out_channels[i - 1]
                if i > 0
                else self.block_out_channels[0]
            )
            is_final_block = i == len(self.block_out_channels) - 1

            if down_block_type == "CrossAttnDownBlock2D":
                down_block = CrossAttnDownBlock2D(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    time_embedding_dim=time_embed_dim,
                    cross_attention_dim=self.cross_attention_dim,
                    num_attention_heads=out_channels // self.attention_head_dim,
                    attention_head_dim=self.attention_head_dim,
                    num_layers=self.layers_per_block,
                    add_downsample=not is_final_block,
                    quant_config=quant_config,
                    prefix=maybe_prefix(prefix, f"down_blocks.{i}"),
                )
            elif down_block_type == "DownBlock2D":
                down_block = DownBlock2D(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    time_embedding_dim=time_embed_dim,
                    num_layers=self.layers_per_block,
                    add_downsample=not is_final_block,
                    prefix=maybe_prefix(prefix, f"down_blocks.{i}"),
                )
            else:
                raise ValueError(
                    f"Unsupported down_block_type: {down_block_type}")

            self.down_blocks.append(down_block)

        # Middle block - UNetMidBlock2DCrossAttn
        mid_block_channel = self.block_out_channels[-1]
        self.mid_block = UNetMidBlock2DCrossAttn(
            in_channels=mid_block_channel,
            time_embedding_dim=time_embed_dim,
            cross_attention_dim=self.cross_attention_dim,
            num_attention_heads=mid_block_channel // self.attention_head_dim,
            attention_head_dim=self.attention_head_dim,
            quant_config=quant_config,
            prefix=maybe_prefix(prefix, "mid_block"),
        )

        # Build up blocks matching Diffusers structure exactly
        self.up_blocks = nn.ModuleList([])

        # up_block_types: ['UpBlock2D', 'CrossAttnUpBlock2D',
        # 'CrossAttnUpBlock2D', 'CrossAttnUpBlock2D']
        up_block_types = config.up_block_types
        reversed_block_out_channels = list(reversed(self.block_out_channels))

        for i, (up_block_type, out_channels) in enumerate(zip(up_block_types, reversed_block_out_channels)):  # noqa: E501
            prev_output_channel = (
                reversed_block_out_channels[i - 1]
                if i > 0
                else mid_block_channel
            )
            is_final_block = i == len(self.block_out_channels) - 1

            if up_block_type == "CrossAttnUpBlock2D":
                up_block = CrossAttnUpBlock2D(
                    in_channels=out_channels,
                    out_channels=out_channels,
                    prev_output_channel=prev_output_channel,
                    time_embedding_dim=time_embed_dim,
                    cross_attention_dim=self.cross_attention_dim,
                    num_attention_heads=out_channels // self.attention_head_dim,  # noqa: E501
                    attention_head_dim=self.attention_head_dim,
                    num_layers=self.layers_per_block + 1,  # 3 for up blocks
                    add_upsample=not is_final_block,
                    quant_config=quant_config,
                    prefix=maybe_prefix(prefix, f"up_blocks.{i}"),
                )
            elif up_block_type == "UpBlock2D":
                up_block = UpBlock2D(
                    in_channels=out_channels,
                    out_channels=out_channels,
                    prev_output_channel=prev_output_channel,
                    time_embedding_dim=time_embed_dim,
                    num_layers=self.layers_per_block + 1,  # 3 for up blocks
                    add_upsample=not is_final_block,
                    prefix=maybe_prefix(prefix, f"up_blocks.{i}"),
                )
            else:
                raise ValueError(f"Unsupported up_block_type: {up_block_type}")

            self.up_blocks.append(up_block)

        # Output layers
        self.conv_norm_out = nn.GroupNorm(
            self.norm_num_groups, self.block_out_channels[0],
        )
        self.conv_out = nn.Conv2d(
            self.block_out_channels[0], self.out_channels,
            kernel_size=3, padding=1,
        )

    def forward(
        self,
        sample: torch.Tensor,
        timestep: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """
        Args:
            sample: (batch_size, in_channels, height, width) noisy samples
            timestep: (batch_size,) timesteps
            encoder_hidden_states: (batch_size, seq_len, cross_attention_dim)
                text embeddings

        Returns:
            (batch_size, out_channels, height, width) predicted noise
        """
        # Time embedding
        timesteps = timestep.flatten()
        t_emb = self.time_proj(timesteps)
        emb = self.time_embedding(t_emb)

        # Input convolution
        sample = self.conv_in(sample)

        # Down blocks - collect skip connections including the initial sample
        # The initial sample is also a skip connection for the final up block
        all_skip_connections = [sample]  # Include initial sample as first skip
        # connection

        for down_block in self.down_blocks:
            if hasattr(down_block, 'forward'):
                # CrossAttnDownBlock2D or DownBlock2D
                if isinstance(down_block, CrossAttnDownBlock2D):
                    sample, res_samples = down_block(
                        sample, emb, encoder_hidden_states,
                    )
                elif isinstance(down_block, DownBlock2D):
                    sample, res_samples = down_block(sample, emb)
                else:
                    raise ValueError(
                        f"Unknown down block type: {type(down_block)}",
                    )

                # Store skip connections for later use
                all_skip_connections.extend(res_samples)

        # Mid block
        sample = self.mid_block(sample, emb, encoder_hidden_states)

        # Up blocks - distribute skip connections properly
        # Skip connections are consumed in LIFO order
        # (last produced, first consumed)
        skip_stack = list(reversed(all_skip_connections))

        for up_block in self.up_blocks:
            # Get the required number of skip connections for this block
            num_res_layers = len(up_block.resnets)

            # Extract skip connections for this block
            block_skip_connections = []
            for _ in range(num_res_layers):
                if skip_stack:
                    block_skip_connections.append(skip_stack.pop(0))
                else:
                    raise ValueError(
                        "Insufficient skip connections for "
                        f"{type(up_block).__name__}",
                    )

            # Convert to tuple and pass to the block
            res_samples = tuple(reversed(block_skip_connections))

            if isinstance(up_block, CrossAttnUpBlock2D):
                sample = up_block(
                    sample, res_samples, emb, encoder_hidden_states,
                )
            elif isinstance(up_block, UpBlock2D):
                sample = up_block(sample, res_samples, emb)
            else:
                raise ValueError(f"Unknown up block type: {type(up_block)}")

        # Output
        sample = self.conv_norm_out(sample)
        sample = F.silu(sample)
        sample = self.conv_out(sample)

        return sample

    def load_weights(self, weights) -> set[str]:
        """Load weights from a diffusers checkpoint."""
        from aphrodite.modeling.models.utils import AutoWeightsLoader

        # Create weight mapping from diffusers to our structure
        weight_mapping = self._create_weight_mapping()

        # Apply weight mapping to the weights
        mapped_weights = []
        for name, loaded_weight in weights:
            mapped_name = weight_mapping.get(name, name)
            mapped_weights.append((mapped_name, loaded_weight))

        # Use AutoWeightsLoader for the actual loading
        loader = AutoWeightsLoader(self)
        return loader.load_weights(mapped_weights)

    def _create_weight_mapping(self) -> dict[str, str]:
        """Create mapping from diffusers weight names to our weight names."""
        mapping = {}

        # Time embedding (direct mapping)
        mapping.update({
            "time_embedding.linear_1.weight": "time_embedding.linear_1.weight",
            "time_embedding.linear_1.bias": "time_embedding.linear_1.bias",
            "time_embedding.linear_2.weight": "time_embedding.linear_2.weight",
            "time_embedding.linear_2.bias": "time_embedding.linear_2.bias",
        })

        # Conv layers (direct mapping)
        mapping.update({
            "conv_in.weight": "conv_in.weight",
            "conv_in.bias": "conv_in.bias",
            "conv_out.weight": "conv_out.weight",
            "conv_out.bias": "conv_out.bias",
            "conv_norm_out.weight": "conv_norm_out.weight",
            "conv_norm_out.bias": "conv_norm_out.bias",
        })

        # Down blocks
        for block_idx in range(4):
            self._map_down_block(mapping, block_idx)

        # Mid block
        self._map_mid_block(mapping)

        # Up blocks
        for block_idx in range(4):
            self._map_up_block(mapping, block_idx)

        return mapping

    def _map_down_block(self, mapping: dict[str, str], block_idx: int):
        """Map down block weights."""
        block_prefix = f"down_blocks.{block_idx}"

        # ResNet blocks (2 per down block)
        for resnet_idx in range(2):
            self._map_resnet_block(
                mapping, f"{block_prefix}.resnets.{resnet_idx}",
            )

        # Attention blocks (only for first 3 down blocks)
        if block_idx < 3:
            for attn_idx in range(2):
                self._map_attention_block(
                    mapping, f"{block_prefix}.attentions.{attn_idx}",
                )

        # Downsampler (except for last block)
        if block_idx < 3:
            mapping.update({
                f"{block_prefix}.downsamplers.0.conv.weight": (
                    f"{block_prefix}.downsamplers.0.conv.weight"
                ),
                f"{block_prefix}.downsamplers.0.conv.bias": (
                    f"{block_prefix}.downsamplers.0.conv.bias"
                ),
            })

    def _map_up_block(self, mapping: dict[str, str], block_idx: int):
        """Map up block weights."""
        block_prefix = f"up_blocks.{block_idx}"

        # ResNet blocks (3 per up block)
        for resnet_idx in range(3):
            self._map_resnet_block(
                mapping, f"{block_prefix}.resnets.{resnet_idx}",
            )

        # Attention blocks (for blocks 1, 2, 3)
        if block_idx > 0:
            for attn_idx in range(3):
                self._map_attention_block(
                    mapping, f"{block_prefix}.attentions.{attn_idx}",
                )

        # Upsampler (except for last block)
        if block_idx < 3:
            mapping.update({
                f"{block_prefix}.upsamplers.0.conv.weight": (
                    f"{block_prefix}.upsamplers.0.conv.weight"
                ),
                f"{block_prefix}.upsamplers.0.conv.bias": (
                    f"{block_prefix}.upsamplers.0.conv.bias"
                ),
            })

    def _map_mid_block(self, mapping: dict[str, str]):
        """Map mid block weights."""
        # 2 ResNet blocks
        for resnet_idx in range(2):
            self._map_resnet_block(mapping, f"mid_block.resnets.{resnet_idx}")

        # 1 Attention block
        self._map_attention_block(mapping, "mid_block.attentions.0")

    def _map_resnet_block(self, mapping: dict[str, str], prefix: str):
        """Map ResNet block weights."""
        mapping.update({
            f"{prefix}.norm1.weight": f"{prefix}.norm1.weight",
            f"{prefix}.norm1.bias": f"{prefix}.norm1.bias",
            f"{prefix}.conv1.weight": f"{prefix}.conv1.weight",
            f"{prefix}.conv1.bias": f"{prefix}.conv1.bias",
            f"{prefix}.time_emb_proj.weight": f"{prefix}.time_emb_proj.weight",
            f"{prefix}.time_emb_proj.bias": f"{prefix}.time_emb_proj.bias",
            f"{prefix}.norm2.weight": f"{prefix}.norm2.weight",
            f"{prefix}.norm2.bias": f"{prefix}.norm2.bias",
            f"{prefix}.conv2.weight": f"{prefix}.conv2.weight",
            f"{prefix}.conv2.bias": f"{prefix}.conv2.bias",
        })

        # Optional shortcut connection (when input/output channels differ)
        mapping.update({
            f"{prefix}.conv_shortcut.weight": f"{prefix}.conv_shortcut.weight",
            f"{prefix}.conv_shortcut.bias": f"{prefix}.conv_shortcut.bias",
        })

    def _map_attention_block(self, mapping: dict[str, str], prefix: str):
        """Map attention block weights (Transformer2DModel)."""
        # Outer projection layers
        mapping.update({
            f"{prefix}.norm.weight": f"{prefix}.norm.weight",
            f"{prefix}.norm.bias": f"{prefix}.norm.bias",
            f"{prefix}.proj_in.weight": f"{prefix}.proj_in.weight",
            f"{prefix}.proj_in.bias": f"{prefix}.proj_in.bias",
            f"{prefix}.proj_out.weight": f"{prefix}.proj_out.weight",
            f"{prefix}.proj_out.bias": f"{prefix}.proj_out.bias",
        })

        # Transformer blocks (usually just 1)
        for transformer_idx in range(1):
            # SD 1.5 has 1 transformer block per attention
            self._map_transformer_block(
                mapping, f"{prefix}.transformer_blocks.{transformer_idx}",
            )

    def _map_transformer_block(self, mapping: dict[str, str], prefix: str):
        """Map BasicTransformerBlock weights."""
        # Layer norms
        mapping.update({
            f"{prefix}.norm1.weight": f"{prefix}.norm1.weight",
            f"{prefix}.norm1.bias": f"{prefix}.norm1.bias",
            f"{prefix}.norm2.weight": f"{prefix}.norm2.weight",
            f"{prefix}.norm2.bias": f"{prefix}.norm2.bias",
            f"{prefix}.norm3.weight": f"{prefix}.norm3.weight",
            f"{prefix}.norm3.bias": f"{prefix}.norm3.bias",
        })

        # Self-attention (attn1)
        mapping.update({
            f"{prefix}.attn1.to_q.weight": f"{prefix}.attn1.to_q.weight",
            f"{prefix}.attn1.to_k.weight": f"{prefix}.attn1.to_k.weight",
            f"{prefix}.attn1.to_v.weight": f"{prefix}.attn1.to_v.weight",
            f"{prefix}.attn1.to_out.0.weight": (
                f"{prefix}.attn1.to_out.0.weight"
            ),
            f"{prefix}.attn1.to_out.0.bias": f"{prefix}.attn1.to_out.0.bias",
        })

        # Cross-attention (attn2)
        mapping.update({
            f"{prefix}.attn2.to_q.weight": f"{prefix}.attn2.to_q.weight",
            f"{prefix}.attn2.to_k.weight": f"{prefix}.attn2.to_k.weight",
            f"{prefix}.attn2.to_v.weight": f"{prefix}.attn2.to_v.weight",
            f"{prefix}.attn2.to_out.0.weight": (
                f"{prefix}.attn2.to_out.0.weight"
            ),
            f"{prefix}.attn2.to_out.0.bias": f"{prefix}.attn2.to_out.0.bias",
        })

        # Feedforward (GeGLU)
        mapping.update({
            f"{prefix}.ff.net.0.proj.weight": (
                f"{prefix}.ff.net.0.0.weight"
            ),  # GeGLU proj layer
            f"{prefix}.ff.net.0.proj.bias": f"{prefix}.ff.net.0.0.bias",
            f"{prefix}.ff.net.2.weight": f"{prefix}.ff.net.2.weight",
            f"{prefix}.ff.net.2.bias": f"{prefix}.ff.net.2.bias",
        })



import argparse

from aphrodite.endpoints.cli.types import CLISubcommand


class BenchmarkSubcommandBase(CLISubcommand):
    """ The base class of subcommands for aphrodite bench. """

    help: str

    @classmethod
    def add_cli_args(cls, parser: argparse.ArgumentParser) -> None:
        """Add the CLI arguments to the parser."""
        raise NotImplementedError

    @staticmethod
    def cmd(args: argparse.Namespace) -> None:
        """Run the benchmark.

        Args:
            args: The arguments to the command.
        """
        raise NotImplementedError



from dataclasses import MISSING, Field, asdict, dataclass, field
from typing import Literal, Union

import pytest

from aphrodite.config import (LoadConfig, ModelConfig, PoolerConfig, AphroditeConfig,
                         config, get_field)
from aphrodite.modeling.layers.pooler import PoolingType
from aphrodite.platforms import current_platform


class TestConfig1:
    pass


@dataclass
class TestConfig2:
    a: int
    """docstring"""


@dataclass
class TestConfig3:
    a: int = 1


@dataclass
class TestConfig4:
    a: Union[Literal[1], Literal[2]] = 1
    """docstring"""


@pytest.mark.parametrize(("test_config", "expected_error"), [
    (TestConfig1, "must be a dataclass"),
    (TestConfig2, "must have a default"),
    (TestConfig3, "must have a docstring"),
    (TestConfig4, "must use a single Literal"),
])
def test_config(test_config, expected_error):
    with pytest.raises(Exception, match=expected_error):
        config(test_config)


def test_get_field():

    @dataclass
    class TestConfig:
        a: int
        b: dict = field(default_factory=dict)
        c: str = "default"

    with pytest.raises(ValueError):
        get_field(TestConfig, "a")

    b = get_field(TestConfig, "b")
    assert isinstance(b, Field)
    assert b.default is MISSING
    assert b.default_factory is dict

    c = get_field(TestConfig, "c")
    assert isinstance(c, Field)
    assert c.default == "default"
    assert c.default_factory is MISSING


@pytest.mark.parametrize(
    ("model_id", "expected_runner_type", "expected_task"),
    [
        ("distilbert/distilgpt2", "generate", "generate"),
        ("intfloat/multilingual-e5-small", "pooling", "embed"),
        ("jason9693/Qwen2.5-1.5B-apeach", "pooling", "classify"),
        ("cross-encoder/ms-marco-MiniLM-L-6-v2", "pooling", "score"),
        ("Qwen/Qwen2.5-Math-RM-72B", "pooling", "reward"),
        ("openai/whisper-small", "transcription", "transcription"),
    ],
)
def test_auto_task(model_id, expected_runner_type, expected_task):
    config = ModelConfig(
        model_id,
        task="auto",
        tokenizer=model_id,
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
    )

    assert config.runner_type == expected_runner_type
    assert config.task == expected_task


@pytest.mark.parametrize(("model_id", "bad_task"), [
    ("Qwen/Qwen2.5-Math-RM-72B", "generate"),
])
def test_incorrect_task(model_id, bad_task):
    with pytest.raises(ValueError, match=r"does not support the .* task"):
        ModelConfig(
            model_id,
            task=bad_task,
            tokenizer=model_id,
            tokenizer_mode="auto",
            trust_remote_code=False,
            seed=0,
            dtype="float16",
        )


MODEL_IDS_EXPECTED = [
    ("Qwen/Qwen1.5-7B", 32768),
    ("mistralai/Mistral-7B-v0.1", 4096),
    ("mistralai/Mistral-7B-Instruct-v0.2", 32768),
]


@pytest.mark.parametrize("model_id_expected", MODEL_IDS_EXPECTED)
def test_disable_sliding_window(model_id_expected):
    model_id, expected = model_id_expected
    model_config = ModelConfig(
        model_id,
        task="auto",
        tokenizer=model_id,
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
        revision=None,
        disable_sliding_window=True,
    )
    assert model_config.max_model_len == expected


def test_get_sliding_window():
    TEST_SLIDING_WINDOW = 4096
    # Test that the sliding window is correctly computed.
    # For Qwen1.5/Qwen2, get_sliding_window() should be None
    # when use_sliding_window is False.
    qwen2_model_config = ModelConfig(
        "Qwen/Qwen1.5-7B",
        task="auto",
        tokenizer="Qwen/Qwen1.5-7B",
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
        revision=None,
    )

    qwen2_model_config.hf_config.use_sliding_window = False
    qwen2_model_config.hf_config.sliding_window = TEST_SLIDING_WINDOW
    assert qwen2_model_config.get_sliding_window() is None

    qwen2_model_config.hf_config.use_sliding_window = True
    assert qwen2_model_config.get_sliding_window() == TEST_SLIDING_WINDOW

    mistral_model_config = ModelConfig(
        "mistralai/Mistral-7B-v0.1",
        task="auto",
        tokenizer="mistralai/Mistral-7B-v0.1",
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
        revision=None,
    )
    mistral_model_config.hf_config.sliding_window = None
    assert mistral_model_config.get_sliding_window() is None

    mistral_model_config.hf_config.sliding_window = TEST_SLIDING_WINDOW
    assert mistral_model_config.get_sliding_window() == TEST_SLIDING_WINDOW


@pytest.mark.skipif(current_platform.is_rocm(),
                    reason="Xformers backend is not supported on ROCm.")
def test_get_pooling_config():
    model_id = "sentence-transformers/all-MiniLM-L12-v2"
    model_config = ModelConfig(
        model_id,
        task="auto",
        tokenizer=model_id,
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
        revision=None,
    )

    pooling_config = model_config._init_pooler_config()
    assert pooling_config is not None

    assert pooling_config.normalize
    assert pooling_config.pooling_type == PoolingType.MEAN.name


@pytest.mark.skipif(current_platform.is_rocm(),
                    reason="Xformers backend is not supported on ROCm.")
def test_get_pooling_config_from_args():
    model_id = "sentence-transformers/all-MiniLM-L12-v2"
    model_config = ModelConfig(model_id,
                               task="auto",
                               tokenizer=model_id,
                               tokenizer_mode="auto",
                               trust_remote_code=False,
                               seed=0,
                               dtype="float16",
                               revision=None)

    override_pooler_config = PoolerConfig(pooling_type='CLS', normalize=True)
    model_config.override_pooler_config = override_pooler_config

    pooling_config = model_config._init_pooler_config()
    assert pooling_config is not None
    assert asdict(pooling_config) == asdict(override_pooler_config)


@pytest.mark.skipif(current_platform.is_rocm(),
                    reason="Xformers backend is not supported on ROCm.")
def test_get_bert_tokenization_sentence_transformer_config():
    bge_model_config = ModelConfig(
        model="BAAI/bge-base-en-v1.5",
        task="auto",
        tokenizer="BAAI/bge-base-en-v1.5",
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
        revision=None,
    )

    bert_bge_model_config = bge_model_config._get_encoder_config()

    assert bert_bge_model_config["max_seq_length"] == 512
    assert bert_bge_model_config["do_lower_case"]


def test_rope_customization():
    TEST_ROPE_SCALING = {"rope_type": "dynamic", "factor": 2.0}
    TEST_ROPE_THETA = 16_000_000.0
    LONGCHAT_ROPE_SCALING = {"rope_type": "linear", "factor": 8.0}

    llama_model_config = ModelConfig(
        "meta-llama/Meta-Llama-3-8B-Instruct",
        task="auto",
        tokenizer="meta-llama/Meta-Llama-3-8B-Instruct",
        tokenizer_mode="auto",
        trust_remote_code=False,
        dtype="float16",
        seed=0,
    )
    assert getattr(llama_model_config.hf_config, "rope_scaling", None) is None
    assert getattr(llama_model_config.hf_config, "rope_theta", None) == 500_000
    assert llama_model_config.max_model_len == 8192

    llama_model_config = ModelConfig(
        "meta-llama/Meta-Llama-3-8B-Instruct",
        task="auto",
        tokenizer="meta-llama/Meta-Llama-3-8B-Instruct",
        tokenizer_mode="auto",
        trust_remote_code=False,
        dtype="float16",
        seed=0,
        hf_overrides={
            "rope_scaling": TEST_ROPE_SCALING,
            "rope_theta": TEST_ROPE_THETA,
        },
    )
    assert getattr(llama_model_config.hf_config, "rope_scaling",
                   None) == TEST_ROPE_SCALING
    assert getattr(llama_model_config.hf_config, "rope_theta",
                   None) == TEST_ROPE_THETA
    assert llama_model_config.max_model_len == 16384

    longchat_model_config = ModelConfig(
        "lmsys/longchat-13b-16k",
        task="auto",
        tokenizer="lmsys/longchat-13b-16k",
        tokenizer_mode="auto",
        trust_remote_code=False,
        dtype="float16",
        seed=0,
    )
    # Check if LONGCHAT_ROPE_SCALING entries are in longchat_model_config
    assert all(
        longchat_model_config.hf_config.rope_scaling.get(key) == value
        for key, value in LONGCHAT_ROPE_SCALING.items())
    assert longchat_model_config.max_model_len == 16384

    longchat_model_config = ModelConfig(
        "lmsys/longchat-13b-16k",
        task="auto",
        tokenizer="lmsys/longchat-13b-16k",
        tokenizer_mode="auto",
        trust_remote_code=False,
        dtype="float16",
        seed=0,
        hf_overrides={
            "rope_scaling": TEST_ROPE_SCALING,
        },
    )
    assert getattr(longchat_model_config.hf_config, "rope_scaling",
                   None) == TEST_ROPE_SCALING
    assert longchat_model_config.max_model_len == 4096


@pytest.mark.skipif(current_platform.is_rocm(),
                    reason="Encoder Decoder models not supported on ROCm.")
@pytest.mark.parametrize(("model_id", "is_encoder_decoder"), [
    ("facebook/opt-125m", False),
    ("facebook/bart-base", True),
    ("meta-llama/Llama-3.2-1B-Instruct", False),
    ("meta-llama/Llama-3.2-11B-Vision", True),
])
def test_is_encoder_decoder(model_id, is_encoder_decoder):
    config = ModelConfig(
        model_id,
        task="auto",
        tokenizer=model_id,
        tokenizer_mode="auto",
        trust_remote_code=False,
        dtype="float16",
        seed=0,
    )

    assert config.is_encoder_decoder == is_encoder_decoder


@pytest.mark.parametrize(("model_id", "uses_mrope"), [
    ("facebook/opt-125m", False),
    ("Qwen/Qwen2-VL-2B-Instruct", True),
])
def test_uses_mrope(model_id, uses_mrope):
    config = ModelConfig(
        model_id,
        task="auto",
        tokenizer=model_id,
        tokenizer_mode="auto",
        trust_remote_code=False,
        dtype="float16",
        seed=0,
    )

    assert config.uses_mrope == uses_mrope


def test_generation_config_loading():
    model_id = "Qwen/Qwen2.5-1.5B-Instruct"

    # When set generation_config to "aphrodite", the default generation config
    # will not be loaded.
    model_config = ModelConfig(model_id,
                               task="auto",
                               tokenizer=model_id,
                               tokenizer_mode="auto",
                               trust_remote_code=False,
                               seed=0,
                               dtype="float16",
                               generation_config="aphrodite")
    assert model_config.get_diff_sampling_param() == {}

    # When set generation_config to "auto", the default generation config
    # should be loaded.
    model_config = ModelConfig(model_id,
                               task="auto",
                               tokenizer=model_id,
                               tokenizer_mode="auto",
                               trust_remote_code=False,
                               seed=0,
                               dtype="float16",
                               generation_config="auto")

    correct_generation_config = {
        "repetition_penalty": 1.1,
        "temperature": 0.7,
        "top_p": 0.8,
        "top_k": 20,
    }

    assert model_config.get_diff_sampling_param() == correct_generation_config

    # The generation config could be overridden by the user.
    override_generation_config = {"temperature": 0.5, "top_k": 5}

    model_config = ModelConfig(
        model_id,
        task="auto",
        tokenizer=model_id,
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
        generation_config="auto",
        override_generation_config=override_generation_config)

    override_result = correct_generation_config.copy()
    override_result.update(override_generation_config)

    assert model_config.get_diff_sampling_param() == override_result

    # When generation_config is set to "aphrodite" and override_generation_config
    # is set, the override_generation_config should be used directly.
    model_config = ModelConfig(
        model_id,
        task="auto",
        tokenizer=model_id,
        tokenizer_mode="auto",
        trust_remote_code=False,
        seed=0,
        dtype="float16",
        generation_config="aphrodite",
        override_generation_config=override_generation_config)

    assert model_config.get_diff_sampling_param() == override_generation_config


@pytest.mark.parametrize("pt_load_map_location", [
    "cuda",
    {
        "": "cuda"
    },
])
def test_load_config_pt_load_map_location(pt_load_map_location):
    load_config = LoadConfig(pt_load_map_location=pt_load_map_location)
    config = AphroditeConfig(load_config=load_config)

    assert config.load_config.pt_load_map_location == pt_load_map_location
